/******************************************************	
	Some code in this file taken/modified from 
	Pandya & Macy 
	Pattern Recognition with Neural Networks in C++
	CRC Press/IEEE Press, 1996
	Thanks to Yooki Park for lending me this book,
	even though it does not receive his stamp of 
	approval!
******************************************************/
//#define MYDEBUG 1		//always inputs the digit 4
//#define KEEPRUNNING	//don't have to press space; reports less

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <math.h>
#include <time.h>
#include "NTadt.h"
#include "rtvideo.h"

#define MAX_RAND 2000

#define MAXLAYERS 5				// Max # of layers in net (in, out, & hidden)
#define MAXNEURONS 170			// Max # of neurons per layer
#define MAXPATTERNS 10000		// Max # of patterns in a training set
#define SIGMOID 0				// Choose squashing fn
#define STEPFN 1				// Choose step fn
#define ARCGRAN 5				// Archive granularity
#define NONE 0					// Do not archive error data
#define ALL 1					// Archive error data for each iteration
#define AVERAGE 2				// Archive average error data for each epoch
#define WORST 3					// Archive worst error for each epoch
#define TRUE 1
#define FALSE 0					
#define SHUFFLE 0				// 0 to defeat random presentation order
								// 1 to randomize presentation of input vectors
FILE* ARCHIVE;



int archOn = AVERAGE;		// Set to ALL, NONE, AVERAGE, or WORST to control
							// data sent to ARCHIVE file for graphical analysis


class BackProp {
private:
	double w[MAXLAYERS][MAXNEURONS][MAXNEURONS];		// Weights matrix
	double wPrev[MAXLAYERS][MAXNEURONS][MAXNEURONS];	// previous weights for momentum
	double neuron[MAXLAYERS][MAXNEURONS];
	double sums[MAXLAYERS][MAXNEURONS];
	double deltaJ[MAXLAYERS][MAXNEURONS];
	double delta_Wij[MAXLAYERS][MAXNEURONS][MAXNEURONS];	
	double error[MAXNEURONS];
	double worstErr;
	double avgErr;
	double lastAvgErr;
	double lastWorstErr;

	//Topology
	//
	int numLayers;				// # of layers
	int outLayerIndx;			// array index of last layer
	int layerSize[MAXLAYERS];	// Number of neurons in each layer

	//Pattern Data
	//
	double inPattern[MAXPATTERNS][MAXNEURONS];	// Input values for each pattern
	double desired[MAXPATTERNS][MAXNEURONS];	// Desired value for each pattern/output
	unsigned int patPresOrder[MAXPATTERNS];		// Order to present patterns to net
	unsigned long int currIter;					// Current iterations number
	long int epoch;
	int numPatterns;							// Total patterns in training set
	int currPat;								// Current pattern used in training
	int convCount;								// Number of consecutive patterns w/in tolerance
	int convergeFlg;							// Flag indicates convergence has occurred

	//Parameters
	//
	double temperature;			// For sigmoid
	double eta;					// Learning rate
	double alpha;				// Momentum
	double errTol;				// Min error tolerance required for convergence
	unsigned long int maxIter;	// Max iterations to do before stopping

public:
	BackProp(void);
	void LoadTrainingSet(char *Fname);
	void GetInputs(void);
	void RunNet(void);
	void GetParms(char*);
	void GetWeights(void);
	double Sigmoid(double Net, double Tempr);
	void SetRandomWeights(void);
	double HCalcDelta(int lyr, int j, double dNetJ);
	int CalcErrors(void);
	void AdaptWeights(void);
	void SaveWeights(char*);
	void WindowStuff(void);
	int Train(char*, char*);
	long int GetEpoch()
		{return epoch;}
	double GetTemperature()
		{return temperature;}
	double GetEta()
		{return eta;}
	double GetAlpha()
		{return alpha;}
	//double* GetWeightsParam()
	//	{return w;}
	int GetLayers()
		{return numLayers;}
	int GetLayerSize(int layer)
		{return layerSize[layer];}

	char selected; //selected to be a parent

	//Graphical display
	GrayImage imLayerZero, imLayerOne, imLayerTwo, 
		imZeroToOne, imOneToTwo;
	drX_cookie *x_cookie;
	drX_windowCookie *w_cookie;
	drX_windowCookie *w2_cookie;
	drX_windowCookie *w3_cookie;
	drX_windowCookie *w4_cookie;
	drX_windowCookie *w5_cookie;


};


// Method definitions for class BackProp:

/*********************************************************
	Constructor - initialize everything
*********************************************************/
BackProp::BackProp()
{
	lastAvgErr = 1.0;
	lastWorstErr = 1.0;
	worstErr = 0.0;
	avgErr = 0.0;
	currIter = 0;
	epoch = 0;
	currPat = -1;		// Current pattern used in training
	convCount = 0;
	convergeFlg = FALSE;
	selected = 0;

	imLayerZero = (GrayImage)imNew(IMAGE_GRAY, 144, 144);
	if (imLayerZero == NULL)
		printf("failed to initialize imLayerZero\n");
	imLayerOne  = (GrayImage)imNew(IMAGE_GRAY, 144, 144);
	imLayerTwo = (GrayImage)imNew(IMAGE_GRAY, 48, 12);
	imZeroToOne = (GrayImage)imNew(IMAGE_GRAY, 144, 144);
	imOneToTwo = (GrayImage)imNew(IMAGE_GRAY, 144, 4);

	// init directx stuff
	x_cookie = drX_init();
	if (x_cookie == NULL){
		printf("failed to initialize directX\n");
		exit(0);
	}
	
	w_cookie = drX_createWindow(x_cookie, 144, 144, "Input layer neurons");
	w2_cookie = drX_createWindow(x_cookie, 144, 144, "Hidden layer neurons");
	w3_cookie = drX_createWindow(x_cookie, 48, 12, "Output layer neurons");
	w4_cookie = drX_createWindow(x_cookie, 144, 144, "0->1 weights");
	w5_cookie = drX_createWindow(x_cookie, 144, 4, "1->2 weights");

}

/*********************************************************
	LoadTrainingSet - takes file representation of
	training set and puts training data & expected
	values into their respective arrays
*********************************************************/

void BackProp::LoadTrainingSet(char* fName)
{
	FILE* pFile;
	int pgIndx;		// neuron index 
	int x, mask;
	int pat, i, j;
	float inVal;
	int numPatBytes;

	pFile = fopen(fName,"r");
	if(pFile == NULL){
		printf("\nUnable to open training set file: %s\n", fName);
		exit(0);
	}
	//printf("\nLoading training set from file: %s\n", fName);

	//fscanf(pFile, "%d", &numPatterns);
	//printf("using default # of patterns - 5000\n");
	numPatterns = 5000;
	numPatBytes = layerSize[0]/8;		// # of input layer neurons must be divisible by 8
		for (pat = 0; pat < numPatterns; pat++){	// For each pattern
		pgIndx = 0;
		for (i=0; i < numPatBytes; i++){		// For each byte in the pattern
			fscanf(pFile, "%x", &x);
			mask = 0x80;						// =128: read first (highest order) bit
			for (j=0; j<8; j++){				// For each bit in the byte
				
				if ((mask & x)>0)
					inPattern[pat][pgIndx]=1.0;
				else
					inPattern[pat][pgIndx]=0.0;

				mask = mask/2;					// read next (lower order) bit
				pgIndx++;
			} 
		} 
		// Get expected values

		for (i=0; i< layerSize[outLayerIndx]; i++) {
			fscanf(pFile, "%f", &inVal);
			desired[pat][i] = inVal;
		} 
	} 
	fclose (pFile);

	// Init pattern presentation order

	for (i=0; i<numPatterns; i++) {
		patPresOrder[i]=i;			//start w/ an unmodified order
	} 
}

/*********************************************************
	GetParms - loads topology from specified file
	if available, else from PARMS1
*********************************************************/
void BackProp::GetParms(char * prmFileName) {
	/*FILE* prmFile;
	prmFile = fopen(prmFileName, "r");
	if (prmFile == NULL) {
		printf("\nUnable to open parameter file: %s \n", prmFileName);
		printf("\nAttempting to open default file: PARMS1");
		prmFile = fopen("PARMS1","r");
		if (prmFile == NULL) {
			printf("\nUnable to open parameter file PARMS1\n"); */
			//printf("using defaults\n");
			temperature = 2;
			eta = 100;
			alpha = 0;
			maxIter = 2000;
			errTol = 0.2;
			numLayers = 3;
			layerSize[0] = 145;
			layerSize[1] = 145;
			layerSize[2] = 4;
		//}
	//}

	/*printf("\nLoading parameters from file: %s\n",prmFileName);
	fscanf(prmFile, "%f", &temperature);
	fscanf(prmFile, "%f", &eta);
	fscanf(prmFile, "%f", &alpha);
	fscanf(prmFile, "%d", &maxIter);
	fscanf(prmFile, "%f", &errTol);
	fscanf(prmFile, "%d", &numLayers);
	printf("\nNumber of layers=%d\n",numLayers);
	for (int i=0; i<numLayers; i++) {
		fscanf(prmFile, "%d", &layerSize[i]);
		printf("Number of neurons in layer %d = %d\n", i, layerSize[i]);
	}*/

	outLayerIndx = numLayers-1;		// Accomodate 0 origined arrays
	//fclose(prmFile);
}

/*********************************************************
	GetWeights - loads a previously saved set of weights
	from specified file.  Weights must match current net
	params.  No checking is done, so watch out!
*********************************************************/
void BackProp::GetWeights()
{
	FILE* wtFile;
	char wtFileName[30];
	int i,j,k;
	double zWT;

	wtFile = fopen("DEFAULT.WGT","r");
	if(wtFile==NULL){
		printf("Unable to open weight file DEFAULT.WGT");
		exit(0);
	}

	for (i=0; i<numLayers-1; i++) 
		for (k=0; k<layerSize[i+1]; k++) 
			for (j=0; j<=layerSize[i]; j++) {
				fscanf(wtFile, "%f", &zWT);
				w[i][j][k] = zWT;
			}

	fclose(wtFile);
}

/*********************************************************
	SetRandomWeights - pretty self explanatory
*********************************************************/
void BackProp::SetRandomWeights(){
	int i,j,k;
	double zWT;
	srand((unsigned)time(NULL));
	for (i=1; i<numLayers; i++)
		for(k=0; k<layerSize[i]; k++)
			for (j=0; j<=layerSize[i-1]; j++){		// <= : One extra for bias neuron
				zWT=(double)rand();
				w[i][j][k]=1-zWT/20000.0;
				//printf ("%f\t",w[i][j][k]);
				wPrev[i][j][k]=0;
			}
}

/*********************************************************
	GetInputs - loads training vectors into input layer
	neurons; keeps track of iteration and epoch
*********************************************************/
void BackProp::GetInputs(){
	int i,j,k,a,b,c;
	unsigned int patID;
	currIter++;				// Update the current iteration counter
	currPat++;				// Update the current pattern counter
	if (currPat>=numPatterns){
		currPat=0;
		epoch++;
	}
	if ((currPat==0) && (SHUFFLE>0)){	//shuffle
		for (j=0; j<=numPatterns; j++) {
			a = numPatterns*(float)rand()/(float)MAX_RAND;
			b = numPatterns*(float)rand()/(float)MAX_RAND;
			c = patPresOrder[a];
			patPresOrder[a] = patPresOrder[b];
			patPresOrder[b]= c;
		}
		if (epoch%100==0){	//every 100 epochs
			printf("Epoch %d: ",epoch);
			printf("Worst=%f Average=%f", lastWorstErr, lastAvgErr);
		}
	}
#ifdef MYDEBUG
	patID = 0;
#else 
	patID = patPresOrder[currPat];
#endif

#ifndef KEEPRUNNING
	printf("examining pattern %d\n",patID);
#endif

	for (i=0; i<layerSize[0]; i++)
		neuron[0][i]=inPattern[patID][i];	// Show it to the neurons
}

/**********************************************************
	RunNet - forward pass - calculates output for all
	network neurons
**********************************************************/
void BackProp::RunNet()
{
	int lyr;		// Layer to calculate
	int dNeuron;	// destination layer neuron
	int sNeuron;	// source layer neuron
	double sumNet;
	double out;
	int dummy;

	for (lyr=1; lyr<numLayers; lyr++){
		neuron[lyr-1][layerSize[lyr-1]]=1.0;	//force bias neuron output to 1.0
		for (dNeuron=0; dNeuron<layerSize[lyr]; dNeuron++){
			sumNet=0.0;
			for(sNeuron=0; sNeuron<=layerSize[lyr-1]; sNeuron++) //<= for bias neuron
				sumNet+=neuron[lyr-1][sNeuron] * w[lyr][sNeuron][dNeuron];
			out = Sigmoid(10-sumNet, temperature);
			//out = Sigmoid(sumNet, temperature);
			neuron[lyr][dNeuron]=out;
			sums[lyr][dNeuron]=sumNet;
		}
	}

	int result = (neuron[2][0]>0.5)*8 + (neuron[2][1]>0.5)*4 
		+ (neuron[2][2]>0.5)*2 + (neuron[2][3]>0.5)*1;

#ifndef KEEPRUNNING
	printf("I think it's %d",result);
	if (((neuron[2][0]>0.2) && (neuron[2][0]<0.8)) ||
		((neuron[2][1]>0.2) && (neuron[2][1]<0.8)) ||
		((neuron[2][2]>0.2) && (neuron[2][2]<0.8)) ||
		((neuron[2][3]>0.2) && (neuron[2][3]<0.8)))
		printf(" , but it's not within tolerance");
#endif
	//scanf("%d",&dummy);
}

/**********************************************************
	HCalcDelta: calculate backward error signal for 
	hidden nodes
**********************************************************/
double BackProp::HCalcDelta(int lyr, int j, double dNETj) {
	int k;
	double delta, sumk;
	sumk=0.0;
	for (k=0; k<layerSize[lyr+1]; k++)
		sumk += deltaJ[lyr+1][k]*w[lyr+1][j][k];
	delta = dNETj*sumk;
	return delta;
}

/**********************************************************
	CalcErrors - determines convergence & archives
	convergence behavior
**********************************************************/
int BackProp::CalcErrors(){
	double dNETj;
	double momentum;
	int i,j,result;
	int localConvFlg = TRUE;
	unsigned int patID;

	// The following line used for DEBUGGING ONLY!!
	//
#ifdef MYDEBUG
	patID = 0;
#else	
	patID = patPresOrder[currPat];
#endif

	result = desired[patID][0] * 8 + 
		desired[patID][1] * 4 + 
		desired[patID][2] * 2 + 
		desired[patID][3] * 1;
	
#ifndef KEEPRUNNING
	printf("\nErrors: I'm trying to adapt the weights to make it %d\n",result);
#endif

	for (j=0; j<layerSize[outLayerIndx]; j++){	// For all the output neurons
		error[j]=(desired[patID][j] - neuron[outLayerIndx][j]);	// Output error
		if (fabs(error[j])>=errTol)
			localConvFlg = FALSE;
		avgErr += fabs(error[j]);
		if (fabs(error[j])>worstErr)
			worstErr=fabs(error[j]);
		if (currPat == (numPatterns-1)) {
			avgErr=avgErr/(double)numPatterns;
			if (archOn == AVERAGE) {
				if ((ARCGRAN*(epoch/ARCGRAN)) == epoch) {	//only save arcGran-th epochs
					fprintf(ARCHIVE,"%d\t%f\t%f\n", epoch, avgErr, fabs(worstErr));
				}
			}
			lastAvgErr = avgErr;
			lastWorstErr = worstErr;
			avgErr = 0.0;
			worstErr = 0.0;
		}
		if (archOn == ALL)
			fprintf(ARCHIVE,"%d\t%f\n", currIter, avgErr, fabs(error[j]));
	}
	if (localConvFlg){
		convCount++;		// Record that another consecutive pattern is within errTol
		if (convCount == 2*numPatterns){
			convergeFlg = TRUE;
			printf("Epoch: %d ", epoch);
			printf("Worst: %f  Avg: %f\n", lastWorstErr, lastAvgErr);
		}
	}
	else
		convCount=0;	// Start over - this pattern had an error out of tolerance

	return convergeFlg;
}

/**********************************************************
	AdaptWeights: back propagation backward pass
**********************************************************/
void BackProp::AdaptWeights()
{
	double dNETj;
	double momentum;
	int lyr, i, j, result;
	unsigned int patID;
	
	//This next line for DEBUGGING ONLY!!
#ifdef MYDEBUG	
	patID=0;
#else
	patID=patPresOrder[currPat];
#endif

	/*result = desired[patID][0] * 8 + 
		desired[patID][1] * 4 + 
		desired[patID][2] * 2 + 
		desired[patID][3] * 1;
	printf("\nAdapt: I'm trying to adapt the weights to make it %d\n",result);*/
	
	//printf("\n\n");
	
	for (lyr=outLayerIndx; lyr>0; lyr--){	// go through each layer, backwards
		for (j=0; j<layerSize[lyr]; j++){	// go through each neuron in layer
			dNETj = neuron[lyr][j]*(1-neuron[lyr][j]);
			//dNETj=((exp(-sums[lyr][j]))/temperature) *
			//	sqrt(Sigmoid(sums[lyr][j],temperature));
			
			
			
			if (lyr == outLayerIndx){
				error[j]=(desired[patID][j] - neuron[lyr][j]);
#ifndef KEEPRUNNING
				printf("Error for neuron %d: %f\n",j,error[j]);
#endif
				deltaJ[lyr][j] = error[j] * dNETj;
			}
			else
				deltaJ[lyr][j] = HCalcDelta(lyr, j, dNETj);
		}
	}
	
	for (lyr = outLayerIndx; lyr>0; lyr--){			// go through each layer, backwards
		for (j=0; j<layerSize[lyr]; j++){			// go through each neuron in layer
			for (i=0; i<= layerSize[lyr-1];i++) {	// go through source (incl. bias) neurons
				delta_Wij[lyr][i][j] = eta * deltaJ[lyr][j] * neuron[lyr-1][i];
				momentum = alpha * (w[lyr][i][j] - wPrev[lyr][i][j]);
				wPrev[lyr][i][j] = w[lyr][i][j];
				w[lyr][i][j] = w[lyr][i][j] + delta_Wij[lyr][i][j] + momentum;
				
				//printf("\nchange of %f",w[lyr][i][j]-wPrev[lyr][i][j]);

			}
		}
	}


}

/***********************************************************
	SaveWeights
***********************************************************/
void BackProp::SaveWeights(char* wgtName)
{
	int lyr, s,d;
	double zWT;
	FILE* weightFile;
	weightFile = fopen(wgtName, "w");
	if (weightFile == NULL){
		printf("Unable to open weight file for output: %s\n", wgtName);
		exit(0);
	}
	printf("Saving calculated weights:\n");
	fprintf(weightFile, "0.00\n");				// Threshold always 0
	fprintf(weightFile, "%f\n", temperature);	// Temperature;
	fprintf(weightFile, "%d\n", numLayers);		// Number of layers in net
	
	for (lyr = 0; lyr<numLayers; lyr++)			// Save topology
		fprintf(weightFile, "%d\n", layerSize[lyr]);	// Number of neurons/layer

	for (lyr = 1; lyr<numLayers; lyr++)				// go through each layer, starting @ 1st hidden
		for (d=0; d<layerSize[lyr]; d++)			// go through each dest. neuron in layer
			for (s=0; s<= layerSize[lyr-1];s++) {	// go through each source neuron
				zWT = w[lyr][s][d];
				fprintf(weightFile, "%f\n", zWT);
			}

	fclose(weightFile);
}

/***********************************************************
	Sigmoid - the activation function
***********************************************************/
double BackProp::Sigmoid(double net, double tempr)
{
	return 1.0/(1.0 + exp (-net/tempr));
}

/***********************************************************
	WindowStuff - create images and output them to windows
***********************************************************/
void BackProp::WindowStuff()
{
	int x,y,i,j;
	
	float max=0;
	float min=9999999;
	
	for (y=0;y<12;y++){
		for (x=0;x<12;x++){
			if (neuron[0][y*12 + x]>max){
				max = neuron[0][y*12 + x];
			}
			if (neuron[0][y*12 + x]<min){
				min = neuron[0][y*12 + x];
			}
		}
	}


	for (y=0; y<12; y++){
		for (x=0; x<12; x++){
			for (j=0; j<12; j++){
				for (i=0; i<12; i++){
					imRef(imLayerZero, x*12 + i, 143-(y*12 + j)) = ((neuron[0][y*12 + x] - min) / (max-min)) * 255;
				}
			}
		}
	}


	max = 0;
	min=9999999;
	for (y=0;y<12;y++)
		for (x=0;x<12;x++){
			if (neuron[1][y*12 + x]>max)
				max = neuron[1][y*12 + x];
			if (neuron[1][y*12 + x]<min)
				min = neuron[1][y*12 + x];
		}


	for (y=0; y<12; y++)
		for (x=0; x<12; x++)
			for (j=0; j<12; j++)
				for (i=0; i<12; i++)
					imRef(imLayerOne,x*12 + i, 143-(y*12 + j)) = ((neuron[1][y*12 + x]-min) / (max-min)) * 255;

	/*max = 0;
	min=9999999;
	for (y=0;y<4;y++){
			if (neuron[2][y]>max)
				max = neuron[2][y];
			if (neuron[2][y]<min)
				min = neuron[2][y];
	}*/


	min = -1;
	max = 1;
	for (x=0; x<4; x++)
			for (j=0; j<12; j++)
				for (i=0; i<12; i++)
					imRef(imLayerTwo, x*12 + i, j) = ((neuron[2][x]-min) / (max-min)) * 255;

	max = 0;
	min=9999999;
	i=1;
		for (y=0;y<layerSize[i-1];y++)
			for (x=0;x<layerSize[i];x++){
				if (w[i][y][x]>max)
					max = w[i][y][x];
				if (w[i][y][x]<min)
					min = w[i][y][x];
			}

	for (y=0;y<144;y++)
		for (x=0;x<144;x++)
			imRef(imZeroToOne, x, 143-y) = ((w[1][x][y]-min) / (max-min)) * 255;

	max = 0;
	min=9999999;
	i=2;
		for (y=0;y<layerSize[i-1];y++)
			for (x=0;x<layerSize[i];x++){
				if (w[i][y][x]>max)
					max = w[i][y][x];
				if (w[i][y][x]<min)
					min = w[i][y][x];
			}


	for (y=0;y<4;y++)
		for (x=0;x<144;x++)
			imRef(imOneToTwo, x, 3-y) = ((w[2][x][y]-min) / (max-min)) * 255;



	

	
	//output to windows
		drX_forceBltGray(w_cookie, imLayerZero);
		drX_forceBltGray(w2_cookie, imLayerOne);
		drX_forceBltGray(w3_cookie, imLayerTwo);
		drX_forceBltGray(w4_cookie, imZeroToOne);
		drX_forceBltGray(w5_cookie, imOneToTwo);
}

/***********************************************************
	Train - top level control to train the network
	with back-propagation
***********************************************************/
int BackProp::Train(char* trnFName, char* parmFName)
{
	int dummy;
	
	if (archOn) ARCHIVE = fopen("archive.lst","w");
	GetParms(parmFName);
	LoadTrainingSet(trnFName);
	SetRandomWeights();
	int converged = 0;
	while ((!converged) && (currIter<maxIter)){
		GetInputs();
		WindowStuff();
		RunNet();
		converged = CalcErrors();
		//printf(" converged = %d\n",converged);
		if ((fabs(error[0])<0.4) && 
			(fabs(error[1])<0.4) && 
			(fabs(error[2])<0.4) &&  
			(fabs(error[3])<0.4))
			printf ("right\n");
		else
			printf("wrong\n");
		AdaptWeights();
#ifndef KEEPRUNNING
		printf("Iteration %d complete\n",currIter);
		printf("Press 0, then enter");
		scanf("%d",&dummy);
#endif
		printf("\n\n");
		
	}

	if (archOn) fclose (ARCHIVE);
	return converged;
}

/***********************************************************
	ShowResults
***********************************************************/
void ShowResults(int convergeFlg, long int currEpoch, double temp, double alpha, double eta)
{
	printf("\n-----------------------------------\n");
	if (convergeFlg)
		printf("Success - convergence has occurred at iteration %d\n", currEpoch);
	else
		printf("Failure - convergence has not occurred.\n");
	printf ("Temperature = %f, ETA = %f, Alpha = %f", temp, eta, alpha);
	printf ("\n----------------------------------\n");

}



BackProp BPnet;		//instantiate the object

int main(int argc, char *argv[]){	
	int converged;
	if (argc<4) {
		printf("Usage: BProp training_file params_file weight_file");
		exit(0);
	}

	
	converged = BPnet.Train(argv[1], argv[2]);

	ShowResults(converged, 
		BPnet.GetEpoch(),
		BPnet.GetTemperature(),
		BPnet.GetAlpha(),
		BPnet.GetEta());

	if (converged)
		BPnet.SaveWeights(argv[3]);
	return 0;
}



 




















