/******************************************************	
	Some code in this file taken/modified from 
	Pandya & Macy 
	Pattern Recognition with Neural Networks in C++
	CRC Press/IEEE Press, 1996
	Thanks to Yooki Park for lending me this book,
	even though it does not receive his stamp of 
	approval!
******************************************************/


// Method definitions for class BackProp:

/*********************************************************
	Constructor - initialize everything
*********************************************************/
BackProp::BackProp()
{
	SetRandomWeights();
	selected = 0;

}


/*********************************************************
	SetRandomWeights - pretty self explanatory
*********************************************************/
void BackProp::SetRandomWeights(){
	int i,j,k;
	double zWT;
	srand((unsigned)time(NULL));
	for (i=1; i<numLayers; i++)
		for(k=0; k<layerSize[i]; k++)
			for (j=0; j<=layerSize[i-1]; j++){		// <= : One extra for bias neuron
				zWT=(double)rand();
				w[i][j][k]=1-zWT/20000.0;
				//printf ("%f\t",w[i][j][k]);
				wPrev[i][j][k]=0;
			}
}

/*********************************************************
	GetInputs - loads training vectors into input layer
	neurons; keeps track of iteration and epoch
*********************************************************/
void BackProp::GetInput(char* fName){
	
	FILE* pFile;
	int pgIndx;		// neuron index 
	int x, currNeuron, mask;
	int pat, i, j;
	float inVal;
	int numPatBytes;

	pFile = fopen(fName,"r");
	if(pFile == NULL){
		printf("\nUnable to open training file: %s\n", fName);
		exit(0);
	}
	
	numPatBytes = layerSize[0]/8;		// # of input layer neurons must be divisible by 8
	currNeuron=0;
	for (i=0; i < numPatBytes; i++){		// For each byte in the pattern
		fscanf(pFile, "%x", &x);
		mask = 0x80;						// =128: read first (highest order) bit
		for (j=0; j<8; j++){				// For each bit in the byte
				
				if ((mask & x)>0)
					neuron[0][currNeuron]=1.0;
				else
					neuron[0][currNeuron]=0.0;

				mask = mask/2;					// read next (lower order) bit
			} 
		} 
		// Get expected values

		for (i=0; i < layerSize[outLayerIndx]; i++) {
			fscanf(pFile, "%f", &inVal);
			desired[i] = inVal;
		} 
	} 
	fclose (pFile);
}

/**********************************************************
	RunNet - forward pass - calculates output for all
	network neurons
**********************************************************/
void BackProp::RunNet()
{
	int lyr;		// Layer to calculate
	int dNeuron;	// destination layer neuron
	int sNeuron;	// source layer neuron
	double sumNet;
	double out;
	int dummy;

	for (lyr=1; lyr<numLayers; lyr++){
		neuron[lyr-1][layerSize[lyr-1]]=1.0;	//force bias neuron output to 1.0
		for (dNeuron=0; dNeuron<layerSize[lyr]; dNeuron++){
			sumNet=0.0;
			for(sNeuron=0; sNeuron<=layerSize[lyr-1]; sNeuron++) //<= for bias neuron
				sumNet+=neuron[lyr-1][sNeuron] * w[lyr][sNeuron][dNeuron];
			out = Sigmoid(10-sumNet, temperature);
			neuron[lyr][dNeuron]=out;
		}
	}

	int result = (neuron[2][0]>0.5)*8 + (neuron[2][1]>0.5)*4 
		+ (neuron[2][2]>0.5)*2 + (neuron[2][3]>0.5)*1;
	printf("I think it's %d",result);
	if (((neuron[2][0]>0.2) && (neuron[2][0]<0.8)) ||
		((neuron[2][1]>0.2) && (neuron[2][1]<0.8)) ||
		((neuron[2][2]>0.2) && (neuron[2][2]<0.8)) ||
		((neuron[2][3]>0.2) && (neuron[2][3]<0.8)))
		printf(" , but it's not within tolerance");
	printf(". Alright?");
	scanf("%d",&dummy);
}

/**********************************************************
	HCalcDelta: calculate backward error signal for 
	hidden nodes
**********************************************************/
double BackProp::HCalcDelta(int lyr, int j, double dNETj) {
	int k;
	double delta, sumk;
	sumk=0.0;
	for (k=0; k<layerSize[lyr+1]; k++)
		sumk += deltaJ[lyr+1][k]*w[lyr+1][j][k];
	delta = dNETj*sumk;
	return delta;
}

/**********************************************************
	CalcErrors - determines convergence & archives
	convergence behavior
**********************************************************/
int BackProp::CalcErrors(){
	double dNETj;
	double momentum;
	int i,j,result;
	int localConvFlg = TRUE;
	unsigned int patID;

	// The following line used for DEBUGGING ONLY!!
	//
	patID = 0;
	
	//patID = patPresOrder[currPat];

	result = desired[patID][0] * 8 + 
		desired[patID][1] * 4 + 
		desired[patID][2] * 2 + 
		desired[patID][3] * 1;
	printf("\nErrors: I'm trying to adapt the weights to make it %d\n",result);
	for (j=0; j<layerSize[outLayerIndx]; j++){	// For all the output neurons
		error[j]=(desired[patID][j] - neuron[outLayerIndx][j]);	// Output error
		if (fabs(error[j])>=errTol)
			localConvFlg = FALSE;
		avgErr += fabs(error[j]);
		if (fabs(error[j])>worstErr)
			worstErr=fabs(error[j]);
		if (currPat == (numPatterns-1)) {
			avgErr=avgErr/(double)numPatterns;
			if (archOn == AVERAGE) {
				if ((ARCGRAN*(epoch/ARCGRAN)) == epoch) {	//only save arcGran-th epochs
					fprintf(ARCHIVE,"%d\t%f\t%f\n", epoch, avgErr, fabs(worstErr));
				}
			}
			lastAvgErr = avgErr;
			lastWorstErr = worstErr;
			avgErr = 0.0;
			worstErr = 0.0;
		}
		if (archOn == ALL)
			fprintf(ARCHIVE,"%d\t%f\n", currIter, avgErr, fabs(error[j]));
	}
	if (localConvFlg){
		convCount++;		// Record that another consecutive pattern is within errTol
		if (convCount == 2*numPatterns){
			convergeFlg = TRUE;
			printf("Epoch: %d ", epoch);
			printf("Worst: %f  Avg: %f\n", lastWorstErr, lastAvgErr);
		}
	}
	else
		convCount=0;	// Start over - this pattern had an error out of tolerance

	return convergeFlg;
}

/**********************************************************
	AdaptWeights: back propagation backward pass
**********************************************************/
void BackProp::AdaptWeights()
{
	double dNETj;
	double momentum;
	int lyr, i, j, result;
	unsigned int patID;
	
	//This next line for DEBUGGING ONLY!!
	patID=0;
	
	//patID=patPresOrder[currPat];

	/*result = desired[patID][0] * 8 + 
		desired[patID][1] * 4 + 
		desired[patID][2] * 2 + 
		desired[patID][3] * 1;
	printf("\nAdapt: I'm trying to adapt the weights to make it %d\n",result);*/
	
	//printf("\n\n");
	
	for (lyr=outLayerIndx; lyr>0; lyr--){	// go through each layer, backwards
		for (j=0; j<layerSize[lyr]; j++){	// go through each neuron in layer
			dNETj=neuron[lyr][j]*(1-neuron[lyr][j]);
			
			
			
			if (lyr == outLayerIndx){
				error[j]=(desired[patID][j] - neuron[lyr][j]);
				
				printf("Error for neuron %d: %f\n",j,error[j]);
				
				deltaJ[lyr][j] = error[j] * dNETj;
			}
			else
				deltaJ[lyr][j] = HCalcDelta(lyr, j, dNETj);
		}
	}
	
	for (lyr = outLayerIndx; lyr>0; lyr--)			// go through each layer, backwards
		for (j=0; j<layerSize[lyr]; j++)			// go through each neuron in layer
			for (i=0; i<= layerSize[lyr-1];i++) {	// go through source (incl. bias) neurons
				delta_Wij[lyr][i][j] = eta * deltaJ[lyr][j] * neuron[lyr-1][i];
				momentum = alpha * (w[lyr][i][j] - wPrev[lyr][i][j]);
				wPrev[lyr][i][j] = w[lyr][i][j];
				w[lyr][i][j] = w[lyr][i][j] + delta_Wij[lyr][i][j] + momentum;
				
				//printf("\nchange of %f",w[lyr][i][j]-wPrev[lyr][i][j]);

			}
}

/***********************************************************
	SaveWeights
***********************************************************/
void BackProp::SaveWeights(char* wgtName)
{
	int lyr, s,d;
	double zWT;
	FILE* weightFile;
	weightFile = fopen(wgtName, "w");
	if (weightFile == NULL){
		printf("Unable to open weight file for output: %s\n", wgtName);
		exit(0);
	}
	printf("Saving calculated weights:\n");
	fprintf(weightFile, "0.00\n");				// Threshold always 0
	fprintf(weightFile, "%f\n", temperature);	// Temperature;
	fprintf(weightFile, "%d\n", numLayers);		// Number of layers in net
	
	for (lyr = 0; lyr<numLayers; lyr++)			// Save topology
		fprintf(weightFile, "%d\n", layerSize[lyr]);	// Number of neurons/layer

	for (lyr = 1; lyr<numLayers; lyr++)				// go through each layer, starting @ 1st hidden
		for (d=0; d<layerSize[lyr]; d++)			// go through each dest. neuron in layer
			for (s=0; s<= layerSize[lyr-1];s++) {	// go through each source neuron
				zWT = w[lyr][s][d];
				fprintf(weightFile, "%f\n", zWT);
			}

	fclose(weightFile);
}

/***********************************************************
	Sigmoid - the activation function
***********************************************************/
double BackProp::Sigmoid(double net, double tempr)
{
	return 1.0/(1.0 + exp (-net/tempr));
}

/***********************************************************
	Train - top level control to train the network
	with back-propagation
***********************************************************/
int BackProp::Train(char* trnFName, char* parmFName)
{
	if (archOn) ARCHIVE = fopen("archive.lst","w");
	GetParms(parmFName);
	LoadTrainingSet(trnFName);
	SetRandomWeights();
	int converged = 0;
	while ((!converged) && (currIter<maxIter)){
		GetInputs();
		RunNet();
		converged = CalcErrors();
		printf(" converged = %d\n",converged); 
		AdaptWeights();
		printf("Iteration %d complete\n",currIter);
	}

	if (archOn) fclose (ARCHIVE);
	return converged;
}

/***********************************************************
	ShowResults
***********************************************************/
void ShowResults(int convergeFlg, long int currEpoch, double temp, double alpha, double eta)
{
	printf("\n-----------------------------------\n");
	if (convergeFlg)
		printf("Success - convergence has occurred at iteration %d\n", currEpoch);
	else
		printf("Failure - convergence has not occurred.\n");
	printf ("Temperature = %f, ETA = %f, Alpha = %f", temp, eta, alpha);
	printf ("\n----------------------------------\n");

}



BackProp BPnet;		//instantiate the object

int main(int argc, char *argv[]){	
	int converged;
	if (argc>3) 
		converged = BPnet.Train(argv[1], argv[2]);
	else {
		printf("Usage: BProp training_file params_file weight_file");
		exit(0);
	}

	ShowResults(converged, 
		BPnet.GetEpoch(),
		BPnet.GetTemperature(),
		BPnet.GetAlpha(),
		BPnet.GetEta());

	if (converged)
		BPnet.SaveWeights(argv[3]);
	return 0;
}



 




















