#ifndef NTADT_H
#define NTADT_H
/*
 *
 *
 * Copyright (c) 1990, 1991, 1992 Cornell University.  All Rights Reserved.  
 *  
 * Copyright (c) 1991, 1992 Xerox Corporation.  All Rights Reserved.  
 *  
 * Use, reproduction, preparation of derivative works, and distribution
 * of this software is permitted.  Any copy of this software or of any
 * derivative work must include both the above copyright notices of
 * Cornell University and Xerox Corporation and this paragraph.  Any
 * distribution of this software or derivative works must comply with all
 * applicable United States export control laws.  This software is made
 * available AS IS, and XEROX CORPORATION DISCLAIMS ALL WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE,
 * AND NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY
 * LIABILITY FOR DAMAGES RESULTING FROM THE SOFTWARE OR ITS USE IS
 * EXPRESSLY DISCLAIMED, WHETHER ARISING IN CONTRACT, TORT (INCLUDING
 * NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION IS ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGES.
 */

/////////////////////////////////////////////////////////////////////////////
// NTadt.h - DLL import specifications for the NTadt library.
//									
// 

/////////////////////////////////////////////////////////////////////////////
// dbug.h
#ifndef TRUE
#define TRUE	(1)
#endif
#ifndef FALSE
#define FALSE	(0)
#endif

#ifndef DEBUG_H
#define DEBUG_H

#ifdef DBUG

/* Need stdio to output to stderr - include it to make sure */
#include <stdio.h>
#define	DEBUG(s, v)	(void)fprintf(stderr, s, v)
#define	VDEBUG(s, v, n)	if (n <= VERBLEVEL) DEBUG(s, v); else

#else

#define	DEBUG(s, v)
#define	VDEBUG(s, v, n)

#endif

// Random number generator stuff.

typedef long GenType;
typedef long GenSeed;

typedef struct _int_generator_struct_ {
  long current;
  long (*next)(struct _int_generator_struct_ *gen);
  GenType type;
  double place_holder;
} *IntGen;

typedef struct _float_generator_struct_ {
  double current;
  double (*next)(struct _float_generator_struct_ *gen);
  GenType type;
} *FloatGen;

typedef struct {long a, b;}             UniformIntParams;
typedef struct {double a, b;}         UniformFloatParams;
typedef struct {double mean, sigma;}      GaussianParams;
typedef struct {double lambda;}        ExponentialParams;

#define UNIFORM_INT_GEN          ((GenType) 1)
#define UNIFORM_FLOAT_GEN        ((GenType) 2)
#define GAUSSIAN_GEN             ((GenType) 3)
#define EXPONENTIAL_GEN          ((GenType) 4)

#define UNIFORM_INT_GEN_TYPE            IntGen
#define UNIFORM_FLOAT_GEN_TYPE        FloatGen
#define GAUSSIAN_GEN_TYPE             FloatGen
#define EXPONENTIAL_GEN_TYPE          FloatGen

#define UNIFORM_INT_GEN_PTYPE           UniformIntParams
#define UNIFORM_FLOAT_GEN_PTYPE       UniformFloatParams
#define GAUSSIAN_GEN_PTYPE                GaussianParams
#define EXPONENTIAL_GEN_PTYPE          ExponentialParams

#define NULLGEN   ((void *) NULL)

#define GEN_RAND_SEED         ((GenSeed) 0)


extern __declspec( dllexport )  void *genNew_(GenType type, GenSeed seed, void *params);
extern __declspec( dllexport )  void  genFree_(GenType type, void *gen);
extern __declspec( dllexport )  void  genReset_(GenType type, void *gen, GenSeed seed);



#define genType(g)                 ((g) ? (g)->type : 0)  
#define genNew(t, s, p)            ({ t ## _PTYPE _p_ = p;    \
				      ((t ## _TYPE) genNew_((t), (s), (void *) &_p_));})
#define genFree(g)                 (genFree_(genType(g), (void *) (g)))
#define genNext(g)                 ((g)->next(g))
#define genCurr(g)                 ((g)->current)
#define genReset(g, s)             (genReset_(genType(g), (void *) (g), (s)))

/////////////////////////////////////////////////////////////////////////////
// priq.h - header file for the priority queue module
typedef struct {
    struct _PriQIntNode *root, *last;
    bool (*compareGT)(void *a, void *b);
    unsigned count;
    } *PriQ;

typedef struct _PriQIntNode {
    struct _PriQIntNode *left, *right, *parent, *next, *prev;
    struct _PriQNode *owner;
    void *userdata;
    } *PriQIntNode;

typedef struct _PriQNode {
    PriQIntNode val;
    PriQ owner;
    } *PriQNode;

/*
 * Constants
 */
#define	NULLPRIQ	((PriQ)NULL)
#define	NULLPRIQNODE	((PriQNode)NULL)

/*
 * Functions
 */
extern __declspec( dllimport ) PriQ pqNew(bool (*compareGT)(void *a, void *b));
extern __declspec( dllimport ) void pqFree(PriQ p);
extern __declspec( dllimport ) PriQNode pqFirst(PriQ p);
extern __declspec( dllimport ) PriQNode pqAdd(PriQ p, void *userdata);
extern __declspec( dllimport ) void pqRem(PriQNode pqn);
extern __declspec( dllimport ) void *pqGet(PriQNode pqn);
extern __declspec( dllimport ) unsigned int pqLen(PriQ p);



/////////////////////////////////////////////////////////////////////////////
// chunk.h - header file for the chunk module
typedef struct {
    struct chunkBit *freelist;
    unsigned size;
	struct hunkBit * allocs;
    } *Chunk;

struct chunkBit {
    struct chunkBit *next;
    };
// wbell 1/17/98
struct hunkBit {
	void * hunk;
	struct hunkBit * next;
	};

/*
 * Constants
 */
#define	NULLCHUNK	((Chunk)NULL)

/*
 * Functions
 */
extern __declspec( dllimport )  Chunk chNew(unsigned size);
extern __declspec( dllimport )  void *chAlloc(Chunk c);
extern __declspec( dllimport )  void chFree(Chunk c, void *p);
extern __declspec( dllimport )  void chNuke(Chunk c);
/*
 * Macros
 */
#define foreach(ln, l) for((ln) = liFirst((l)); (ln) != NULLLISTNODE; (ln) = liNext((ln)))
#define forafter(ln1, ln2) for ((ln1) = liNext(ln2); (ln1) != NULLLISTNODE; (ln1) = liNext((ln1)))


/////////////////////////////////////////////////////////////////////////////
// adt_strings.h - header file for the chunk module
typedef struct {
  char *data;
  int mxlen;
} adtStringHeader;

typedef adtStringHeader *adtString;

#define ADT_NULLSTRING       ((adtString) NULL)

extern __declspec( dllimport )  adtString  strNew(int mxlen);
extern __declspec( dllimport )  void    strFree(adtString str);
extern __declspec( dllimport )  adtString  strClear(adtString str);
extern __declspec( dllimport )  adtString  strNewCopy(adtString str);
extern __declspec( dllimport )  adtString  strMake(char *s);
extern __declspec( dllimport )  int     strLen(adtString str);
extern __declspec( dllimport )  int     strCmp(adtString a, adtString b);
extern __declspec( dllimport )  int     strMaxLen(adtString str);
extern __declspec( dllimport )  char   *strGet(adtString str);
extern __declspec( dllimport )  char    strRef(adtString str, int i);
extern __declspec( dllimport )  adtString  strInit_(adtString *dest, char *s);
#define strInit(dest, s)          strInit_(&(dest), (s))
extern __declspec( dllimport )  adtString  strCopy_(adtString *dest, adtString src);
#define strCopy(dest, src)        strCopy_(&(dest), (src))
extern __declspec( dllimport )  adtString  strCat_(adtString *dest, adtString src);
#define strCat(dest, src)         strCat_(&(dest), (src))
extern __declspec( dllimport )  adtString  strAddChar_(adtString *dest, char ch);
#define strAddChar(dest, ch)      strAddChar_(&(dest), (ch))
extern __declspec( dllimport )  int     strNewLen_(adtString *dest, int newlen);
#define strNewLen(dest, newlen)   strNewLen_(&(dest), (newl))


/////////////////////////////////////////////////////////////////////////////
// misc.h - miscellaneous types and declarations
/* Stuff needed for gregk */
typedef unsigned char uchar;

#ifndef SVR4
typedef unsigned long ulong;
#endif

typedef unsigned int uint;

typedef struct {uchar r, g, b;} RGB;
typedef struct {float r, g, b;} RGBFloat;
typedef struct {float h, s, v;} HSV;
#define import         extern __declspec( dllimport ) 
#define export
#define local          static

/*
 * Stuff for C*; these are snitched from their stdlib.h.
 * I don't include their stdlib.h because it declares malloc as returning
 * char *, which conflicts with the nice definition below.
 */
#ifdef	cstar
extern __declspec( dllimport )  void psrand(unsigned seed);
extern __declspec( dllimport )  int:current prand(void);
extern __declspec( dllimport )  void deallocate_shape(shape *s);
extern __declspec( dllimport )  void:void * palloc(shape s, int bsize);
extern __declspec( dllimport )  void pfree(void:void *pvar);

/* They don't implement volatile, so we get lots of warnings... */
#define	volatile
#endif

#ifndef __GNUC__
#ifdef	cstar

/* Use the C* extensions: >? (MAX) and <? (MIN) operators */
#ifndef	MAX
#define	MAX(x, y) ((x) >? (y))
#endif

#ifndef	MIN
#define	MIN(x, y) ((x) <? (y))
#endif

#ifndef	ABS
#define	ABS(x)	((x) >? (-(x)))
#endif

#else

#ifndef MAX
#define MAX(x, y) (((x) > (y)) ? (x) : (y))
#endif

#ifndef MIN
#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#endif

#ifndef ABS
#define ABS(x) (((x) < 0) ? -(x) : (x))
#endif

/* Added from gregk's misc.h */
#ifndef SWAP
#define SWAP(a, b)  ({ typeof (a) _tmp = (a); (a) = (b); (b) = _tmp; })
#endif

#endif

#else

/* GCC allows new scopes inside expressions; use to get better MAX and MIN */

#define MAX(a, b)   ({ typeof (a) _a = (a); typeof (b) _b = (b); _a > _b ? _a : _b; })
#define MIN(a, b)   ({ typeof (a) _a = (a); typeof (b) _b = (b); _a < _b ? _a : _b; })
#define ABS(a)      ({ typeof (a) _a = (a); _a < (typeof (a)) 0 ? -_a : _a; })
#define SQUARE(x)   ({ typeof (x) _x = (x); _x*_x; })
#define SWAP(a, b)  ({ typeof (a) _tmp = (a); (a) = (b); (b) = _tmp; })
#define ROUND(a)    ({ typeof (a) __a = (a); (int) ((__a > 0) ? (__a + 0.5) : (__a - 0.5)); })
#define FLOAT_NAN     ({ float f; (*((long *) &f) = 0x7fffffff); f;})
#define DOUBLE_NAN    ({ double d; (*((long *) &d) = 0x7fffffff); (*(((long *) &d) + 1) = 0xffffffff); d;})


#endif

#define BOUND(x, l, h)   (MIN(MAX((x), (l)), (h)))

/* This is what malloc returns on error, and is also a null pointer */
#ifndef NULL
#define NULL    0
#endif

/*
 * Include common files. These come AFTER the defines above,
 * because mine are better.
 */
#include <stdio.h>
#ifndef SVR4
#include <stdlib.h>
#endif
#include <time.h>
#include <string.h>
#include <memory.h>
// #include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
// #include <sys/socket.h>
// #include <sys/time.h>
// #include <sys/ipc.h>
// #include <sys/shm.h>
// #include <sys/sem.h>
#include <float.h>

/* Some library routines that never seem to get declared */

/* I don't feel like including setjmp.h, so this is commented out for now */
/*
extern __declspec( dllimport )  int _setjmp(jmp_buf env);
extern __declspec( dllimport )  volatile void _longjmp(jmp_buf env, int val);
*/

/* String functions: string[s].h doesn't have prototypes for all of them. */
/* Who needs prototypes? This breaks on SVR4 and I don't see why they 
 * should be here in the first place. Commented out until someone complains.
 * --mdw 
 */
/*
 * extern __declspec( dllimport )  int strcasecmp(char *s1, char *s2);
 * extern __declspec( dllimport )  int strncasecmp(char *s1, char *s2, int n);
 * extern __declspec( dllimport )  char *strdup(char *s);
 */


/* Sigh... index is a MACRO! */
/* 
 * extern __declspec( dllimport )  char *index(char *s, int c);
 */

/* Raw IO functions */
/* 
 * extern __declspec( dllimport )  int ioctl(int fd, unsigned int request, void *arg);
 * extern __declspec( dllimport )  int truncate(char *path, long length);
 * extern __declspec( dllimport )  int select(int width, fd_set *readfds, fd_set *writefds,
 *		  fd_set *exceptfds, struct timeval *timeout);
 * extern __declspec( dllimport )  int flock(int fd, int operation);
 * extern __declspec( dllimport )  caddr_t mmap(caddr_t addr, size_t len, int prot, int flags, int fd,
 *		    off_t off);
 * extern __declspec( dllimport )  int munmap(caddr_t addr, int len);
 */

/* Socket-fiddling functions */
// extern __declspec( dllimport )  int socket(int domaint, int type, int protocol);
// extern __declspec( dllimport )  int connect(int s, struct sockaddr *name, int namelen);
// extern __declspec( dllimport )  int bind(int s, struct sockaddr *name, int namelen);
// extern __declspec( dllimport )  int listen(int s, int backlog);
// extern __declspec( dllimport )  int accept(int s, struct sockaddr *addr, int *addrlen);

// #ifndef SVR4
/* Random stuff */
//  extern __declspec( dllimport )  long random(void);
// #endif
//  extern __declspec( dllimport )  void srandom(int seed);

// extern __declspec( dllimport )  double drand48();
// extern __declspec( dllimport )  long lrand48();
// extern __declspec( dllimport )  void srand48(long seedval);

/* Time stuff */

// char *strptime(char *buf, char *fmt, struct tm *tm);
// time_t timegm(struct tm *tm);
// time_t timelocal(struct tm *tm);
// void tzset(void);
// void tzsetwall(void);

#ifndef SVR4
/* Memory fiddling routines */
// extern __declspec( dllimport )  void bcopy(char *b1, char *b2, int length);
// extern __declspec( dllimport )  void bzero(char *b, int length);
#endif
 
/* Shared memory stuff */
/*
 * extern __declspec( dllimport )  int shmget(key_t key, int size, int shmflg);
 * extern __declspec( dllimport )  void *shmat(int shmid, void *shmaddr, int shmflg);
 * extern __declspec( dllimport )  int shmdt(void *shmaddr);
 * extern __declspec( dllimport )  int shmctl(int shmid, int cmd, struct shmid_ds *buf);
 */

/* Semaphore stuff */
/* extern __declspec( dllimport )  int semop(int semid, struct sembuf *sops, int nsops);
 * extern __declspec( dllimport )  int semget(key_t key, int nsems, int semflg);
 * extern __declspec( dllimport )  int semctl(int semid, int semnum, int cmd, union semun arg);
 */

/* and misc stuff */
/* extern __declspec( dllimport )  int getopt(int argc, char **argv, char *optstring);
 */
// extern __declspec( dllimport )  char *optarg;
// extern __declspec( dllimport )  int optind, opterr;
/* extern __declspec( dllimport )  void perror(const char *s);
 * extern __declspec( dllimport )  int gethostname(char *name, int namelen);
 * extern __declspec( dllimport )  int usleep(int time);
 */

/* 
 * extern __declspec( dllimport )  int fprintf(FILE *stream, const char *format, ...);
 * extern __declspec( dllimport )  int printf(const char *format, ...);
 * extern __declspec( dllimport )  int fscanf(FILE *stream, const char *format, ...);
 * extern __declspec( dllimport )  int sscanf(const char *s, const char *format, ...);
 * 
 * extern __declspec( dllimport )  int _filbuf();
 * extern __declspec( dllimport )  int _flsbuf();
 * extern __declspec( dllimport )  int fclose(FILE *stream);
 * extern __declspec( dllimport )  int fread(char *ptr, unsigned size, unsigned nitems, FILE *stream);
 * extern __declspec( dllimport )  int fwrite(const char *ptr, unsigned size, unsigned nitems, FILE *stream);
 * extern __declspec( dllimport )  int fflush(FILE *stream);
 * extern __declspec( dllimport )  int ungetc(int c, FILE *stream);
 */

#endif


/* For FILE */
#include <stdio.h>
#include <io.h>

#include <math.h>

inline int round(float x){
		return (fabs(x-(float)((int)x))<.5 ? ((int) x) : ((int)x+(x>0? 1:-1)));
	};

#ifdef	IMAGE_CHECKALL
/* Need eassert() and panic() */
//#include "panic.h"
#endif

/* The number of colours per component (or total for gray images) */
#define	COLDEPTH	8
#define COLRNG		(1 << COLDEPTH)

typedef unsigned short ImageType;

/* The image types */
#define	IMAGE_GRAY	((ImageType) (1 <<  0))
#define	IMAGE_FLOAT	((ImageType) (1 <<  1))
#define	IMAGE_RGB	((ImageType) (1 <<  2))
#define	IMAGE_DOUBLE	((ImageType) (1 <<  3))
#define	IMAGE_BINARY	((ImageType) (1 <<  4))
#define	IMAGE_LONG	((ImageType) (1 <<  5))
#define	IMAGE_PTR	((ImageType) (1 <<  6))
#define	IMAGE_SHORT	((ImageType) (1 <<  7))
#define IMAGE_HSV       ((ImageType) (1 <<  8))
#define IMAGE_RGBF      ((ImageType) (1 <<  9))
#define IMAGE_STRUCT    ((ImageType) (1 << 10))

#define IMAGE_ANY       (IMAGE_GRAY   | IMAGE_FLOAT | IMAGE_RGB  | IMAGE_DOUBLE |      \
			 IMAGE_BINARY | IMAGE_LONG  | IMAGE_PTR  | IMAGE_SHORT  |      \
			 IMAGE_HSV    | IMAGE_RGBF  | IMAGE_STRUCT)


struct ImageHeader;

#define DEF_IMG(imgname, elemtype) \
  typedef elemtype * imgname ## PixPtr; \
  typedef struct imgname ## ImageStruct { \
    elemtype **data; \
    struct ImageHeader *header; \
    struct {void *ptr;} userdata; \
  } * imgname ## Image;

  DEF_IMG(Gray,     uchar);
  DEF_IMG(Float,    float);
  DEF_IMG(RGB,      RGB);
  DEF_IMG(Double,   double);
  DEF_IMG(Binary,   char);
  DEF_IMG(Long,     long);
  DEF_IMG(Ptr,      void *);
  DEF_IMG(Short,    short);
  DEF_IMG(HSV,      HSV);
  DEF_IMG(RGBFloat, RGBFloat);

  DEF_IMG(Any,    void);		     /* generic image */

  typedef unsigned char adtBOOL;
  
#undef DEF_IMG

typedef struct ImageHeader {		     /* An invisible image header type */
  int width, height;
  int xbase, ybase;
  ImageType tag;
  short elemsize, headsize;
  AnyImage anyim;
  adtString docstring;
  adtBOOL wrapped;
  void *Ximage;				     /* only x-stuff knows about this */
} ImageHeader;



/* and some manipulation routines */
extern __declspec( dllimport ) void *im_Load(ImageType types, char *filename);
extern __declspec( dllimport )  void *im_LoadH(ImageType types, int handle);

extern __declspec( dllimport )  void im_Free_(void *im, ImageType type);
extern __declspec( dllimport )  void *im_LoadInto_(void *im, ImageHeader *header, char *filename);
extern __declspec( dllimport )  void *im_LoadHInto_(void *im, ImageHeader *header, int handle);


/************************************************************************************/
extern __declspec( dllimport )  int im_Save_(void *im, ImageHeader *header, char *filename);
/************************************************************************************/

extern __declspec( dllimport )  int im_SaveH_(void *im, ImageHeader *header, int handle);
extern __declspec( dllimport )  void *im_Dup_(void *im, ImageHeader *header);
extern __declspec( dllimport )  void im_Init_(void *im, ImageHeader *header, void *val);
extern __declspec( dllimport )  void im_SetOffset_(void *im, ImageHeader *header, int xbase, int ybase);
extern __declspec( dllimport )  void im_Apply_(void *f, void *im, ImageHeader *header);


/************************************************************************************/
extern __declspec( dllimport )  void *im_NewOrWrap_(ImageType type, int width, int height, int xbase, int ybase,
			   void *raw, int elemsize, int headsize, void *dupimg);
/************************************************************************************/

extern __declspec( dllimport )  void im_MaxAndMin_(void *im, ImageHeader *header, void *max, void *min);
extern __declspec( dllimport )  void im_MaxAndMinNoInf_(void *im, ImageHeader *header, void *maxp, void *minp);
extern __declspec( dllimport )  void *im_Crop_(void *im, ImageHeader *header, int xbase, int ybase, int w, int h);
extern __declspec( dllimport )  void *im_Crop_Into_(void *im, ImageHeader *header, void *outimage, 
				int in_x, int in_y, int out_x, int out_y,
				int w, int h);
extern __declspec( dllimport )  AnyImage imBorders(AnyImage wo, int borderSize);

/* and some useful routines -- some are purposely *not* lvalues */
//#ifndef	IMAGE_CHECKALL
// GROUP_START "Image Access"

// Used to get the width of an image, in pixels
// <b>i</b> is an image
/* EXAMPLE
  GrayImage im = imLoad(IMAGE_GRAY, filename);
  int width = imGetWidth(im);
  */
#define	imGetWidth(i)	((i)->header->width)

// Used to get the height of an image, in pixels
// <b>i</b> is an image
/* EXAMPLE
  GrayImage im = imLoad(IMAGE_GRAY, filename);
  int height = imGetHeight(im);
  */
#define	imGetHeight(i)	((i)->header->height)

// Gets the x value for where the image begins
//  (where it is offset to)
#define	imGetXBase(i)	((i)->header->xbase)

// Gets the y value for where the image begins
//  (where it is offset to)
#define	imGetYBase(i)	((i)->header->ybase)

// Gets the type of image 
// (eg. IMAGE_GRAY)
#define imGetType(i)	((i)->header->tag)

#define imDocString(i)  ((i)->header->docstring)
#define imUserData(i)   ((i)->userdata)
//#else
//#define	imGetWidth(i)	({ assert((i) != NULL); (i)->header->width; })
//#define	imGetHeight(i)	({ assert((i) != NULL); (i)->header->height; })
//#define	imGetXBase(i)	({ assert((i) != NULL); (i)->header->xbase; })
//#define	imGetYBase(i)	({ assert((i) != NULL); (i)->header->ybase; })
//#define imGetType(i)	({ assert((i) != NULL); (i)->header->tag; })
//#define imDocString(i)  (eassert((i) != NULL), ((i)->header->docstring))
//#define imUserData(i)   (eassert((i) != NULL), ((i)->userdata))
//#endif
			 
// Gets the base of the array that is the image data
// Since a 2D image is actually stored as a 1D array this will 
//  return a pointer to that array
// This is useful if you want to iterative through the entire image
//  in linear fashion (like to do a histogram).
// One problem with this is that since it uses imRef you can't pass in 
//  an AnyImage as <b>i</b>.  To work around it you can get the base of the
//  image array by accessing it directly (thorugh structure members).
#define imGetStore(i)		(&(imRef((i), imGetXBase(i), imGetYBase(i))))

// Gets the length of the image data (= width * height)
#define imGetStoreLen(i)	(imGetWidth(i) * imGetHeight(i))

#define imGetXMax(i)       (imGetXBase(i) + imGetWidth(i)  - 1)
#define imGetYMax(i)       (imGetYBase(i) + imGetHeight(i) - 1)
#define imAnyImage(i)      ((i) ? (i)->header->anyim : (AnyImage) NULL)

// GROUP_END
/* define a new structure image type */
#define imDefStructImage(IMGNAME, ELEMTYPE, USERDATA)                                  \
  typedef struct {                                                                     \
    ELEMTYPE **data;                                                                   \
    ImageHeader *header;                                                               \
    USERDATA userdata;                                                                 \
  } *IMGNAME

typedef struct {void *ptr;} StdImUserData;   /* standard image user data field */

#define imNewStruct(imtype, w, h)     imNewOffsetStruct(imtype, (w), (h), 0, 0)
/*
#define imNewOffsetStruct(imtype, w, h, x, y)                                          \
     ({ imtype __i;                                                                    \
	(imtype) im_NewOrWrap_(IMAGE_STRUCT, (w), (h), (x), (y), NULL,                 \
		      sizeof(typeof(**(__i->data))), sizeof(typeof(*__i)), NULL);})
*/

#define imNULL ((AnyImage) NULL)
		
/*
 * These macros are strange - they attempt to make imFree etc somewhat
 * polymorphic by figuring out, in a context where the type is known (at
 * compile-time), the type of the image and passing that in. If it was
 * not done this way, imFree etc would not be able to figure out what
 * sort of Image they had been passed (since the structures may not match).
 *
 * imFree has imGetType and so implicitly checks for (i) != NULL
 */
//GROUP_START "Image Construction-Destruction"
#define imFree(i)	   { if ((void *) (i) != (void *) NULL) {                     \
				 im_Free_((void *)(i), imGetType((i)));                \
			         (i) = NULL;}}

#define imNew(t, w, h)                 imNewOffset((t), (w), (h), 0, 0)
#define imNewOffset(t, w, h, x, y)     im_NewOrWrap_((t), (w), (h), (x), (y),          \
						     NULL, 0, 0, NULL)
#define imWrap(r, t, w, h)             imWrapOffset((r), (t), (w), (h), 0, 0)
#define imWrapOffset(r, t, w, h, x, y) im_NewOrWrap_((t), (w), (h), (x), (y),          \
						     (void *) (r), 0, 0, NULL)
#define imLoad(t, f) im_Load((t),(f))
#define imLoadF(t,f) im_LoadH((t),_fileno(f))
#define imLoadH(t,h) im_LoadH((t),(h))

#ifndef	IMAGE_CHECKALL
#define imLoadInto(i, n)   im_LoadInto_((void *)(i), ((i)->header), (n))
#define imLoadFInto(i, f)  im_LoadHInto_((void *)(i), ((i)->header), _fileno(f))
#define imLoadHInto(i, h)  im_LoadHInto_((void *)(i), ((i)->header), (h))
#define imSave(i, n)	   im_Save_((void *)(i), ((i)->header), (n))
#define imSaveF(i, f)	   im_SaveH_((void *)(i), ((i)->header), _fileno(f))
#define imSaveH(i, h)	   im_SaveH_((void *)(i), ((i)->header), h)	
#define	imDup(i)	   im_Dup_((void *)(i), ((i)->header))

// this gives compiler probelms under nt
// #define imInit(i, val)     { typeof (**((i)->data)) __value = (val);                    \
//			      im_Init_((void *)(i), ((i)->header), (void *) &__value);})

#define imInit(i, val) { im_Init_((void *)i, ((i)->header), (void *) &val); }
#define imSetOffset(i, x, y) im_SetOffset_((void *)(i), ((i)->header), (x), (y))
#define imApply(func, i)   { void                                                     \
			      (*__func)(v, int x, int y) = func; \
			      im_Apply_((void *) __func, (void *)(i), ((i)->header));}





// this gives compiler probelms under nt
/*
#define imMaxAndMin(i, max, min)   ({ typeof (**((i)->data)) __max, __min;             \
					im_MaxAndMin_((void *)(i), ((i)->header),      \
						      (void *)&__max, (void *)&__min); \
					(min) = (typeof (min)) __min;                  \
					(max) = (typeof (max)) __max;})
*/
// this gives compiler probelms under nt
/*
#define imMaxAndMinNoInf(i, max, min)   ({ typeof (**((i)->data)) __max, __min;        \
					im_MaxAndMinNoInf_((void *)(i), ((i)->header), \
						      (void *)&__max, (void *)&__min); \
					(min) = (typeof (min)) __min;                  \
					(max) = (typeof (max)) __max;})
*/					
// these don't work right
/*
#define imCrop(im, xb, yb, w, h) im_Crop_((void *) (im), ((im)->header), \
							 (xb), (yb), (w), (h)))

#define imCropInto(in, out, in_x, in_y, out_x, out_y, w, h)  \
				im_Crop_Into_((void *) (in), ((in)->header), \
                               (void *) (out), \
			       (in_x),(in_y),(out_x),(out_y),(w),(h))) 
*/
#else
/*
#define imLoadInto(i, n)   (eassert((i) != NULL),                                      \
			    (im_LoadInto_((void *)(i), ((i)->header), (n))))
#define imLoadFInto(i, f)  (eassert((i) != NULL),                                      \
			    (im_LoadFInto_((void *)(i), ((i)->header), (f))))
#define imSave(i, n)	   (eassert((i) != NULL),                                      \
			    (im_Save_((void *)(i), ((i)->header), (n))))
#define imSaveF(i, f)	   (eassert((i) != NULL),                                      \
			    (im_SaveF_((void *)(i), ((i)->header), (f))))
#define	imDup(i)	(eassert((i) != NULL), (im_Dup_((void *)(i), ((i)->header))))

// this gives compiler probelms under nt
#define imInit(i, val)  (eassert((i) != NULL),                                         \
			 ({ typeof (**((i)->data)) __value = (val);                    \
			    im_Init_((void *)(i), ((i)->header), (void *) &__value);}))
				
#define imSetOffset(i, x, y) (eassert((i) != NULL),                                    \
			    (im_SetOffset_((void *)(i), ((i)->header), (x), (y))))

// this gives compiler probelms under nt
#define imApply(func, i)   (eassert((i) != NULL), ({ void                              \
			      (*__func)(typeof (*((i)->data)) v, int x, int y) = func; \
			      im_Apply_((void *) __func, (void *)(i), ((i)->header));}))

// this gives compiler probelms under nt
#define imMaxAndMin(i, max, min)   (eassert((i) != NULL),                              \
				    ({ typeof (**((i)->data)) __max, __min;            \
					 im_MaxAndMin_((void *)(i), ((i)->header),     \
						       (void *)&__max, (void *)&__min); \
					 (min) = (typeof (min)) __min;                 \
					 (max) = (typeof (max)) __max;}))

// this gives compiler probelms under nt
#define imMaxAndMinNoInf(i, max, min)   (eassert((i) != NULL),                         \
				    ({ typeof (**((i)->data)) __max, __min;            \
					 im_MaxAndMinNoInf_((void *)(i), ((i)->header), \
						       (void *)&__max, (void *)&__min); \
					 (min) = (typeof (min)) __min;                 \
					 (max) = (typeof (max)) __max;}))

#define imCrop(im, xb, yb, w, h) im_Crop_((void *) (im), ((im)->header),\
							  (xb), (yb), (w), (h))))

// XXX imCropInto doesn't check anything right now... mdw 
#define imCropInto(in, out, in_x, in_y, out_x, out_y, w, h)  \
  ((typeof (in)) im_Crop_Into_((void *) (in), ((in)->header), \
                               (AnyImage *)(out), \
			       (in_x),(in_y),(out_x),(out_y),(w),(h)))
*/
#endif
// GROUP_END
#define  imBoundsCheck(i, x, y)    (imBoundsCheckX((i), (x))  &&                   \
                                    imBoundsCheckY((i), (y)))
#define  imBoundsCheckX(i, x)      (imBoundsCheckXMax((i), (x)) &&                 \
                                    imBoundsCheckXMin((i), (x)))
#define  imBoundsCheckY(i, y)      (imBoundsCheckYMax((i), (y)) &&                 \
                                    imBoundsCheckYMin((i), (y)))
#define  imBoundsCheckXMax(i, x)   ((x) <= imGetXMax((i)))
#define  imBoundsCheckXMin(i, x)   ((x) >= imGetXBase((i)))
#define  imBoundsCheckYMax(i, y)   ((y) <= imGetYMax((i)))
#define  imBoundsCheckYMin(i, y)   ((y) >= imGetYBase((i)))

/* and an access routine - gives an *lvalue* */
#ifndef	IMAGE_CHECKALL
// GROUP_START "Image Access"

// Used to reference pixel at (x,y) in image i
// EXAMPLE:
//  int val = imRef(i, x, y);
//  val++;
//  imRef(i, x, y) = val;
#define	imRef(i, x, y)	(((i)->data)[(y)][(x)])

// GROUP_END
#else
static char __imPanicStr[100];
#define	imRef(i, x, y)	(eassert(i != NULL),                                       \
			 ((((x) < (i)->header->xbase) ||                           \
			   ((y) < (i)->header->ybase) ||                           \
			   ((x) >= (i)->header->xbase + (i)->header->width) ||     \
			   ((y) >= (i)->header->ybase + (i)->header->height))      \
			  ? ({sprintf(__imPanicStr,                                \
				      "image bounds check fail: "                  \
				      "%d <= %d < %d, %d <= %d < %d",              \
				      (i)->header->xbase,                          \
				      (x),                                         \
				      (i)->header->xbase + (i)->header->width,     \
				      (i)->header->ybase,                          \
				      (y),                                         \
				      (i)->header->ybase + (i)->header->height);   \
			       panic(__imPanicStr);}), ((i)->data[0][0])           \
			  : (((i)->data)[(y)][(x)])))
#endif





/* Here come macros for increased efficiency access */

#ifndef IMAGE_CHECKALL

// GROUP_START "Image Access"

// Gets a pointer to a pixel (as opposed to a reference)
#define imGetPixPtr(i, x, y) (&(imRef((i), (x), (y))))

// gives the reference from a pointer
// <b>pp</b> is a pixel pointer
#define imPtrRef(i, pp) (*(pp))

// duplicates a pointer to a pixel
// <b>pp</b> is a pixel pointer
#define imPtrDup(i, pp) (pp)

// Moves <b>pp</b> up one pixel in the image
// WARNING: doesn't do bounds checking
// <b>pp</b> is a pixel pointer
#define imPtrUp(i, pp) ((pp) -= (i)->header->width)

// Moves <b>pp</b> down one pixel in the image
// WARNING: doesn't do bounds checking
// <b>pp</b> is a pixel pointer
#define imPtrDown(i, pp) ((pp) += (i)->header->width)

// Moves <b>pp</b> left one pixel in the image
// WARNING: doesn't do bounds checking
// <b>pp</b> is a pixel pointer
#define imPtrLeft(i, pp) ((pp)--)

// Moves <b>pp</b> right one pixel in the image
// WARNING: doesn't do bounds checking
// <b>pp</b> is a pixel pointer
#define imPtrRight(i, pp) ((pp)++)

// compares two pixel pointers to see if they point to
//  the same pixle.
// NOTE: meant for use within same image
// <b>pp</b> is a pixel pointer
#define imPtrEq(i, pp1, pp2) ((pp1) == (pp2))
// GROUP_END



#else

/* Check a pointer against bounds; eval expr if successful */
#define __IMPCHECK(i, pp, expr) (eassert((i) != NULL), ((((pp) < (&((i)->data[(i)->header->ybase][(i)->header->xbase]))) || ((pp) > &((i)->data[(i)->header->yba se+(int)(i)->header->height-1][(i)->header->xbase+(int)(i)->header->width-1]))) ? (({ panic("PixPtr bounds check fail"); }), (expr)) : (expr)))

/* imRef is error-checked already */
#define imGetPixPtr(i, x, y) (&(imRef((i), (x), (y))))
#define imPtrRef(i, pp) __IMPCHECK((i), (pp), (*(pp)))
#define imPtrDup(i, pp) __IMPCHECK((i), (pp), (pp))
#define imPtrUp(i, pp) __IMPCHECK((i), (pp), ((pp) -= (i)->header->width))
#define imPtrDown(i, pp) __IMPCHECK((i), (pp), ((pp) += (i)->header->width))
#define imPtrLeft(i, pp) __IMPCHECK((i), (pp), ((pp)--))
#define imPtrRight(i, pp) __IMPCHECK((i), (pp), ((pp)++))
#define imPtrEq(i, pp1, pp2) __IMPCHECK((i), (pp1), __IMPCHECK((i), (pp2), ((pp1) == (pp2))))


#endif


// ImAdditions.h
__declspec( dllimport )  AnyImage imBorders(AnyImage wo, int borderSize);
//		Creates a new image identical to the input but with a border of borderSize pixels on each side
//		of the original image.
__declspec( dllimport )  AnyImage imCrop(AnyImage im, int topX, int topY, int bottomX, int bottomY);
//		Creates a new image that is the cropped rectangle (topX,topY) (bottomX,bottomY)
//		out of the input.
__declspec( dllimport )  AnyImage imCropInto(AnyImage into, AnyImage from, 
											 int top, int left, int bottom, int right,
											 int toleft, int totop);
//		Similar to imCrop, but instead of creating a new image, it puts the cropped
//		rectangle in into, at the coordinates (toleft,totop)
__declspec( dllimport )  AnyImage imScale(AnyImage input, int newWidth, int newHeight);
//		Creates a new image that is the input scaled to the new width and height
__declspec( dllimport )  AnyImage imDupInto(AnyImage into, AnyImage in);
//		Given 2 images of the same size/type, this copies the contents of in to into into


#endif
