
#include <stdlib.h>
#include <time.h>
#include "MyNet.h"

#define HOWMANY 50			// How many individuals in the population
#define MAX 1.1				// Expected value of highest (#N) ranked indiv.
#define MIN 2-MAX			// Expected value of lowest (#1) ranked indiv.
#define MUTATIONDELTA 0.1	// Max change in weight due to mutation
#define MAXPATTERNS 10000	// Max # of input test patterns
#define TESTEACH 25			//# times each indiv. is tested to quantify fitness
#define GOODENOUGH 80

#define RANDOMIZE srand((unsigned)time(NULL))
#define RANDFLOAT ((double)rand())/((double)MAXRAND))

typedef struct 
{
	float fitness;
	int index;
}  fit_idx;

//Comparison function used by qsort to rank individuals
//By fitness
//
int Compare (fit_idx *arg1, fit_idx *arg2)
{
	if (arg1->fitness < arg2->fitness)
		return -1;
	if (arg1->fitness == arg2->fitness)
		return 0;
	return 1;
}

//Distance in 4-bit representation of 2 numbers
int myDist(int one, int other)
{
	int val=0;

	val += abs((one & 1) - (other & 1));
	val += abs((one & 2) - (other & 2));
	val += abs((one & 4) - (other & 4));
	val += abs((one & 8) - (other & 8));

	return val;
}










class Population {
private:
	MyNet individual[HOWMANY];
	MyNet offspring[HOWMANY];
	fit_idx fitnesses[HOWMANY];
	void Mate(void);
	void Crossover(int, int, int);
	void Mutate(void);
	void InitFitness(int);
	void Select(void);
	void GiveInputs(int,int);

	double inPattern[MAXPATTERNS][MAXNEURONS];
	double desired[MAXPATTERNS][MAXNEURONS];

public:
	Population(char*);
	void SetFitnesses();
	void AdvanceGeneration(void);
	
	float bestPercentage;
	int generation;
	
}

Population::Population(const char* fName)
{
	int i;
	FILE* pFile;
	int pgIndx;		//neuron index
	int x, mask;
	int pat, j;
	float inVal;
	int numPatBytes;
	
	generation = 0;
	bestPercentage = 0;
	
	//Create the individuals in the population
	for (i=0; i<HOWMANY; i++){
		fitnesses[i]->index = i;
		fitnesses[i]->fitness = 0;
		
		individual[i] = new myNet;
		individual[i]->SetRandomWeights;
	}

	//Create the objects that will hold the data of the offspring
	for (i=0; i<HOWMANY; i++){
		offspring[i] = new myNet;

	//Load up the samples
	pFile = fopen(fName,"r");
	if(pFile == NULL){
		printf("\nUnable to open training set file: %s\n", fName);
		exit(0);
	}
	printf("\nLoading training set from file: %s\n", fName);

	printf("using default # of patterns - 5000\n");
	numPatterns = 5000;
	numPatBytes = individual[0]->GetLayerSize(0)/8;		// # of input layer neurons must be divisible by 8
		for (pat = 0; pat < numPatterns; pat++){	// For each pattern
		pgIndx = 0;
		for (i=0; i < numPatBytes; i++){		// For each byte in the pattern
			fscanf(pFile, "%x", &x);
			mask = 0x80;						// =128: read first (highest order) bit
			for (j=0; j<8; j++){				// For each bit in the byte
				
				if ((mask & x)>0)
					inPattern[pat][pgIndx]=1.0;
				else
					inPattern[pat][pgIndx]=0.0;

				mask = mask/2;					// read next (lower order) bit
				pgIndx++;
			} 
		} 
		// Get expected values

		for (i=0; i< layerSize[outLayerIndx]; i++) {
			fscanf(pFile, "%f", &inVal);
			desired[pat][i] = inVal;
		} 
	} 
	fclose (pFile);

}


Population::GiveInputs(int indiv, int whichInput)
{
	for (int i = 0 ; (i<individual[indiv]->GetLayerSize(0))-1 ; i++)
		individual[indiv]->SetNeuron(0,i,inPattern[whichInput][i]);
}

void Population::InitFitness(int thisIndiv)
{
	RANDOMIZE;
	
	int whichPat, rep;

	fitnesses[thisIndiv]->fitness = 0;

	for(rep=0; rep<TESTEACH; rep++){

		whichPat = RANDFLOAT * 5000;

		GiveInputs(thisIndiv, whichPat);
		individual[thisIndiv]->RunNet;

		fitnesses[thisIndiv]->fitness += 4 - myDist(individual[thisIndiv]->Output(), desired[whichPat]);

	}

	if (fitnesses[thisIndiv]->fitness / (4*TESTEACH))
		bestPercentage = fitnesses[thisIndiv]->fitness;

}
	

Population::AdvanceGeneration()
{
	//SetFitnesses();
	Select();
	Mate();
	Mutate();
	SetFitnesses();
	generation++;
}

Population::SetFitnesses()
{
	
	for (int i=0; i<HOWMANY; i++){
		InitFitness(i);
		(fitnesses[i])->idx = i;
	}

	//sort by fitness
	qsort((void *)fitnesses, HOWMANY, sizeof(fit_idx),Compare);

}

Population::Select()
{
	RANDOMIZE;
	
	float sumExpVals = 0;

	//Set expected values to implement Rank selection
	for (int i=0; i<HOWMANY; i++){
		(fitnesses[i]).fitness = 
			MIN + (MAX - MIN)*((i-1)/(HOWMANY-1));
		sumExpVals += (fitnesses[i]).fitness; 
	}
	
	float randomFlt;
	int indivIdx;
	float localSum;

	//Roulette wheel selection
	for (i=0;i<HOWMANY;i++){
		
		
		//gets a random number between 0 and HOWMANY
		randomFlt = RANDFLOAT*HOWMANY;
		
		indivIdx=-1;
		localSum=0;
		while (sumExpVals < randomFlt) {
			indivIdx++;
			localSum += (fitnesses[indivIdx]).fitness;
		}

		(individual[indivIdx]).selected = 1;
	}
}

Population::Mate()
{
	int parent1, parent2;
	float randFlt;

	RANDOMIZE;

	for (int i=0; i<HOWMANY; i++){

		randFlt = RANDFLOAT;
		parent1 = (int)(randFlt*HOWMANY);

		randFlt = RANDFLOAT;
		parent2 = (int)(randFlt*HOWMANY);

		//choose offpring[i]'s parents, randomly
		while ((individual[parent1]->selected==0) &&
			(individual[parent2]->selected==0)) {
			if individual[parent1]->selected==0{
				randFlt = RANDFLOAT;
				parent1 = (int)(randFlt*HOWMANY);
			}
			if individual[parent2]->selected==0{
				randFlt = RANDFLOAT;
				parent2 = (int)(randFlt*HOWMANY);
			}
		}

		Crossover(parent1, parent2, i);
	}

	individual=offspring;	//The offspring become the new generation
}


/***********************************************************************
	Crossover: implements two-point crossover.  The genotype is the array
	of weights.  Each weight-layer experiences 2-pt. crossover independent 
	of the other layer.  For crossing over, the members of each 
	(inter-)layer's 2-d weight-array (source x destination) are considered
	linearly, that is, weight[layer][source][lastdest] is followed by
	weight[layer][source+1][firstdest].
***********************************************************************/
void Population::Crossover(int parent1, int parent2, int destChild)
{
	RANDOMIZE;
	
	int oneCrossS, twoCrossS, oneCrossD, twoCrossD, tempS, tempD;
	int i, j, k;

	int thisLayerSize, prevLayerSize;

	for (i=1; i<3; i++){	//for each layer

		thisLayerSize = individual[0]->GetLayerSize(i);
		prevLayerSize = individual[0]->GetLayerSize(i-1);

		oneCrossS = (int) (RANDFLOAT * (thisLayerSize-1));
		oneCrossD = (int) (RANDFLOAT * (thisLayerSize-1));
		twoCrossS = (int) (RANDFLOAT * (thisLayerSize-1));
		twoCrossD = (int) (RANDFLOAT * (thisLayerSize-1));

		
		//Make sure the start x-over point is before the end x-over pt
		if ((oneCrossS * thisLayerSize + oneCrossD) > 
			(twoCrossS * thisLayerSize + twoCrossD)){
			tempS = oneCrossS;
			tempD = oneCrossD;
			oneCrossS = twoCrossS;
			oneCrossD = twoCrossD;
			twoCrossS = tempS;
			twoCrossD = tempD;
		}

		//switch the segments interior/exterior to [startCross..endCross]
		//for the offspring
		for (j=0; j<prevLayerSize; j++){
			for (k=0; k<thisLayerSize; k++){
				if ((j*thisLayerSize + k)  <  (oneCrossS*thisLayerSize + oneCrossD))
					offspring[destChild]->SetWeight(i,j,k,individual[parent1]->GetWeight(i,j,k));
				else if ((j*thisLayerSize + k) > (twoCrossS*thisLayerSize + twoCrossD))
					offspring[destChild]->SetWeight(i,j,k,individual[parent1]->GetWeight(i,j,k));
				else
					offspring[destChild]->SetWeight(i,j,k,individual[parent2]->GetWeight(i,j,k));

			}
		}
	} //for each layer
}


void Population::Mutate(){}	//For now - blank; many GA's don't use mutation






Population thePopulation;	// Instantiate the class

void main()
{
	Population thePopulation("training");
	thePopulation.SetFitnesses();

	while (thePopulation.bestPercentage < GOODENOUGH)
		thePopulation.AdvanceGeneration();

	printf("GA produced a suitable individual in generation %d\n.", thePopulation.generation);
}











					

	
