/* Library of commands to command the Sony Pantilt Camera EVI-D30 */
#ifndef PAN_TILT_H
#define PAN_TILT_H

#include "..\rtvideo.h"

#define ON 0
#define OFF 1
#define	NORMALMODE 0
#define ATMODE 1
#define MDMODE 2
#define ENTRYMODE1  0
#define ENTRYMODE2  1	
#define ENTRYMODE3  2     
#define ENTRYMODE4  3
#define AUTO	    0
#define INDOORMODE	1
#define OUTDOORMODE	2
#define ONEPUSHMODE	3
#define MANUAL	    1	
#define SHUTTERPRIORITY	2
#define IRISPRIORITY	3
#define BRIGHTCONTROL   4
#define NTSC 0
#define PAL 1
#define REFRESHMODE1 0
#define REFRESHMODE2 1
#define REFRESHMODE3 2

typedef struct {
	HANDLE hCom;	// handle to com port
	char unit;		// 'address' of unit we are going to talk to (use 8 for broadcast)
	BOOL executing_command;		// whether or not the camera is moving or whatever
} pantilt_cookie;


/* to initialize a cookie using COM1, and talking to camera 1, have the following line of code:
	
	pantilt_cookie *cookie;
	if ((cookie = pantilt_init("COM1",1)) == NULL) {
		goto bailout;
	}
This creates a cookie to be used to send all commands*/
/*at end use pantilt_close(cookie) to close connection to the camera*/



RTV_DECL 
pantilt_cookie *pantilt_init(char *com, char unit);/* creates function that initializes the communications*/

RTV_DECL int pantilt_close(pantilt_cookie *cookie);/*closes connection to com*/
RTV_DECL int pantilt_finish_command(pantilt_cookie *cookie);/*checks to see if there is a command being executed and waits until there is no command*/
RTV_DECL int pantilt_receive(pantilt_cookie *cookie, unsigned char *buffer, unsigned int size);/*reads the first 6 bytes of the com buffer, 




/**move functions**/
/* tell camera to move.  pan_speed= 01 to 24;  tilt_speed= 01 to 20; pan_pos= FC90 to 0370; tilt_pos= FED4 to 012C
limits pan= -880 to 880(-100degrees to 1000degrees)
	   tilt= -300 to 300(-25degrees to 25degrees)
real max speep pan=80degrees/sec
			   tilt=50degrees/sec    */

/*these commands block the camera until the cammera has stoped moving or a pantilt_cancel is called,
most command are not executed until these are complete.  Ifclear, almost stops these commands*/
RTV_DECL int pantilt_abspos(pantilt_cookie *cookie, unsigned char pan_speed, unsigned char tilt_speed,
				   short pan_pos, short tilt_pos);/* movesto an absolute position in the grid of pan_pos and tilt_pos*/
RTV_DECL int pantilt_relpos(pantilt_cookie *cookie, unsigned char pan_speed, unsigned char tilt_speed,
				   short pan_pos, short tilt_pos);/*adds pan_pos and tilt_pos to the current position of the camera*/

/*once these commands are called they will send the ack and completion messega right after the camera recieves the message.
So the camera will not block any other subsequent commands*/
RTV_DECL int pantilt_home(pantilt_cookie *cookie);
RTV_DECL int pantilt_reset(pantilt_cookie *cookie);
RTV_DECL int pantilt_up(pantilt_cookie *cookie, unsigned char tilt_speed);
RTV_DECL int pantilt_down(pantilt_cookie *cookie, unsigned char tilt_speed);
RTV_DECL int pantilt_left(pantilt_cookie *cookie, unsigned char pan_speeed);
RTV_DECL int pantilt_right(pantilt_cookie *cookie, unsigned char pan_speeed);
RTV_DECL int pantilt_upleft(pantilt_cookie *cookie, unsigned char pan_speeed, unsigned char tilt_speed);
RTV_DECL int pantilt_upright(pantilt_cookie *cookie, unsigned char pan_speeed, unsigned char tilt_speed);
RTV_DECL int pantilt_downleft(pantilt_cookie *cookie, unsigned char pan_speeed, unsigned char tilt_speed);
RTV_DECL int pantilt_downright(pantilt_cookie *cookie, unsigned char pan_speeed, unsigned char tilt_speed);
RTV_DECL int pantilt_stop(pantilt_cookie *cookie);

// takes negative or positive velocities and calls the correct functions
RTV_DECL void pantilt_velocity(pantilt_cookie *p_cookie, int xvel, int yvel);

/**cancel**/

RTV_DECL int pantilt_cancel(pantilt_cookie *cookie, unsigned char socket);/*cancels some commands, socket= 0 or 1*/

/* zoom */
RTV_DECL int pantilt_zoom_stop(pantilt_cookie *cookie);
RTV_DECL int pantilt_zoom_telestd(pantilt_cookie *cookie);
RTV_DECL int pantilt_zoom_widestd(pantilt_cookie *cookie);
RTV_DECL int pantilt_zoom_televar(pantilt_cookie *cookie,unsigned char telespeed);/*telespeed= 2(low speed) to 7(high speed)*/
RTV_DECL int pantilt_zoom_widevar(pantilt_cookie *cookie,unsigned char telespeed);
RTV_DECL int pantilt_zoom_direct(pantilt_cookie *cookie,short directzoom);/* directzoom= 0000(wide) to 03ff(1023,tele)*/


/*power*/  
/*If the power swithch in the back of the camera is on, 
then the camera can be turned on and off using this command*/
RTV_DECL int power_on(pantilt_cookie *cookie);
RTV_DECL int power_off(pantilt_cookie *cookie);

/*focus*/
RTV_DECL int pantilt_focus_stop(pantilt_cookie *cookie);
RTV_DECL int pantilt_focus_far(pantilt_cookie *cookie);
RTV_DECL int pantilt_focus_near(pantilt_cookie *cookie);
RTV_DECL int pantilt_focus_onauto(pantilt_cookie *cookie);
RTV_DECL int pantilt_focus_onmanual(pantilt_cookie *cookie);
RTV_DECL int pantilt_focus_automanual(pantilt_cookie *cookie);
RTV_DECL int pantilt_focus_direct(pantilt_cookie *cookie, unsigned short directzoom);/*directzoome=0 to ffff*/

/**limits*///sets limits to the motion of the camera
RTV_DECL int pantilt_setlim_upri(pantilt_cookie *cookie,  short pan_pos, short tilt_pos);
RTV_DECL int pantilt_setlim_downle(pantilt_cookie *cookie,  short pan_pos, short tilt_pos);
RTV_DECL int pantilt_clearlim_upri(pantilt_cookie *cookie); 
RTV_DECL int pantilt_clearlim_downle(pantilt_cookie *cookie); 



/*white balance*/
RTV_DECL int pantilt_wb_auto(pantilt_cookie *cookie);
RTV_DECL int pantilt_wb_indoor(pantilt_cookie *cookie);
RTV_DECL int pantilt_wb_outdoor(pantilt_cookie *cookie);
RTV_DECL int pantilt_wb_onepush(pantilt_cookie *cookie);
RTV_DECL int pantilt_wb_onepushtrigger(pantilt_cookie *cookie);

/*Exposure*/
RTV_DECL int pantilt_ae_auto(pantilt_cookie *cookie);
RTV_DECL int pantilt_ae_manual(pantilt_cookie *cookie);
RTV_DECL int pantilt_ae_shutterp(pantilt_cookie *cookie);
RTV_DECL int pantilt_ae_irisp(pantilt_cookie *cookie);
RTV_DECL int pantilt_ae_bright(pantilt_cookie *cookie);

/*brightness*/
RTV_DECL int pantilt_bright_reset(pantilt_cookie *cookie);
RTV_DECL int pantilt_bright_up(pantilt_cookie *cookie);
RTV_DECL int pantilt_bright_down(pantilt_cookie *cookie);

/*shutter speed*/
RTV_DECL int pantilt_shutter_reset(pantilt_cookie *cookie);
RTV_DECL int pantilt_shutter_up(pantilt_cookie *cookie);
RTV_DECL int pantilt_shutter_down(pantilt_cookie *cookie);
RTV_DECL int pantilt_shutter_direct(pantilt_cookie *cookie, unsigned char shutter);/*shutter:00(1/60) to 1B(1/10000sec.)*/ 

/* Iris */
RTV_DECL int pantilt_iris_reset(pantilt_cookie *cookie);
RTV_DECL int pantilt_iris_up(pantilt_cookie *cookie);
RTV_DECL int pantilt_iris_down(pantilt_cookie *cookie);
RTV_DECL int pantilt_iris_direct(pantilt_cookie *cookie, unsigned char iris);/* Iris:00(close) to 11(F1.8)*/

/*gain*/
RTV_DECL int pantilt_gain_reset(pantilt_cookie *cookie);
RTV_DECL int pantilt_gain_up(pantilt_cookie *cookie);
RTV_DECL int pantilt_gain_down(pantilt_cookie *cookie);
RTV_DECL int pantilt_gain_direct(pantilt_cookie *cookie, unsigned char gain);/*gain:00(0 dB) to 07(+18 dB)*/

/* back light*/
RTV_DECL int pantilt_backl_on(pantilt_cookie *cookie); 
RTV_DECL int pantilt_backl_off(pantilt_cookie *cookie);

/* set positions*/
RTV_DECL int pantilt_preset_reset(pantilt_cookie *cookie, unsigned char posi);/*posi:0(position 1) to 5 (position 6)*/
RTV_DECL int pantilt_preset_set(pantilt_cookie *cookie, unsigned char posi);/*posi:0(position 1) to 5 (position 6)*/
RTV_DECL int pantilt_preset_recall(pantilt_cookie *cookie, unsigned char posi);/*posi:0(position 1) to 5 (position 6)*/

/*key lock*/
RTV_DECL int pantilt_keylock_off(pantilt_cookie *cookie);
RTV_DECL int pantilt_keylock_on(pantilt_cookie *cookie);

/*IR recieve, not recieve IR*/
RTV_DECL int pantilt_ir_receive_on(pantilt_cookie *cookie);
RTV_DECL int pantilt_ir_receive_off(pantilt_cookie *cookie);
RTV_DECL int pantilt_ir_receive_onoff(pantilt_cookie *cookie);

/* IR recieve return,  when receive IR, output the data*/
/*when on, the camera will output things whenever the remote cotrol is used*?
/*THIS FUNCTION SHOULD BE OFF, since it messes up pantilt_receive because of the unexpected output*/   
RTV_DECL int pantilt_ir_rcvrtn_on(pantilt_cookie *cookie);
RTV_DECL int pantilt_ir_rcvrtn_off(pantilt_cookie *cookie);



/* wide con lens set */
RTV_DECL int pantilt_wide_conlensset(pantilt_cookie *cookie, unsigned char sensitivity);//when using wide con lens, compensats AT
																			   // sensitivity=0(not using) to 7(x0.6)

/* Datascreen */
RTV_DECL int pantilt_datascreen_on(pantilt_cookie *cookiey);
RTV_DECL int pantilt_datascreen_off(pantilt_cookie *cookiey);
RTV_DECL int pantilt_datascreen_onoff(pantilt_cookie *cookiey);


/*AT mode*/
RTV_DECL int pantilt_atmode_on(pantilt_cookie *cookie);
RTV_DECL int pantilt_atmode_off(pantilt_cookie *cookie);
RTV_DECL int pantilt_atmode_onoff(pantilt_cookie *cookie);

/*AT AE*/
RTV_DECL int pantilt_atae_onoff(pantilt_cookie *cookie);
RTV_DECL int pantilt_atae_onoff(pantilt_cookie *cookie);
RTV_DECL int pantilt_atae_onoff(pantilt_cookie *cookie);

/*AT/MD Frame display*/
RTV_DECL int pantilt_atmd_on(pantilt_cookie *cookie);
RTV_DECL int pantilt_atmd_off(pantilt_cookie *cookie);
RTV_DECL int pantilt_atmd_onoff(pantilt_cookie *cookie);

/*AT/MD Start/Stop*/
RTV_DECL int pantilt_atmd_startstop(pantilt_cookie *cookie);


/*AT offset*/
RTV_DECL int pantilt_atoffset_on(pantilt_cookie *cookie);
RTV_DECL int pantilt_atoffset_off(pantilt_cookie *cookie);
RTV_DECL int pantilt_atoffset_onoff(pantilt_cookie *cookie);

/* AT chase*/
RTV_DECL int pantilt_atchase1(pantilt_cookie *cookie); 
RTV_DECL int pantilt_atchase2(pantilt_cookie *cookie);
RTV_DECL int pantilt_atchase3(pantilt_cookie *cookie); 


/* AT entry*/
RTV_DECL int pantilt_atentry1(pantilt_cookie *cookie);
RTV_DECL int pantilt_atentry2(pantilt_cookie *cookie);
RTV_DECL int pantilt_atentry3(pantilt_cookie *cookie);
RTV_DECL int pantilt_atentry4(pantilt_cookie *cookie);



/**** Motion detector functions ****/
/*Motion Detector mode, on and off functions*/
RTV_DECL int pantilt_mdmode_on(pantilt_cookie *cookie);
RTV_DECL int pantilt_mdmode_off(pantilt_cookie *cookie);
RTV_DECL int pantilt_mdmode_onoff(pantilt_cookie *cookie);


RTV_DECL int pantilt_mdframe(pantilt_cookie *cookie); 
RTV_DECL int pantilt_mddetect(pantilt_cookie *cookie);
RTV_DECL int pantilt_atlinfo(pantilt_cookie *cookie);
RTV_DECL int pantilt_mdlinfo(pantilt_cookie *cookie);

/*MD adjust,  adjust= 0 to F*/
RTV_DECL int pantilt_mdadjust_y(pantilt_cookie *cookie, unsigned char adjust);
RTV_DECL int pantilt_mdadjust_hue(pantilt_cookie *cookie, unsigned char adjust);
RTV_DECL int pantilt_mdadjust_size(pantilt_cookie *cookie, unsigned char adjust);
RTV_DECL int pantilt_mdadjust_dtime(pantilt_cookie *cookie, unsigned char adjust);
RTV_DECL int pantilt_mdadjust_refresh1(pantilt_cookie *cookie);
RTV_DECL int pantilt_mdadjust_refresh2(pantilt_cookie *cookie);
RTV_DECL int pantilt_mdadjust_refresh3(pantilt_cookie *cookie);
RTV_DECL int pantilt_mdadjust_reftime(pantilt_cookie *cookie, unsigned char adjust);


/* Sets adress */
RTV_DECL int pantilt_adress(pantilt_cookie *cookie); 


/*auto tracing zoom*/
RTV_DECL int pantilt_atautozoom_on(pantilt_cookie *cookie);
RTV_DECL int pantilt_atautozoom_off(pantilt_cookie *cookie);
RTV_DECL int pantilt_atautozoom_onoff(pantilt_cookie *cookie);


/*keyboard induced movements*/
RTV_DECL int takekey(pantilt_cookie *cookie);//moves while arrows are pressed down. Zoom F1(tele) F2(wide) F3(stop zoom)
RTV_DECL int stepmove(pantilt_cookie *cookie, unsigned char pan_speed, unsigned char tilt_speed,  
			     short xstep, short ystep);// moves with arrows but with assigned steps
RTV_DECL int waitforclear(pantilt_cookie *cookie);// waits until keyboard is clear








/*** Inquiry ***/
/* see definitions on top of page for return values*/
/*for the status, refer to manual of the camera*/
RTV_DECL int pantilt_inq_posi(pantilt_cookie *cookie, short *pan, short *tilt);


RTV_DECL int pantilt_inq_focus(pantilt_cookie *cookie, unsigned short *focus);


RTV_DECL int pantilt_inq_zoom(pantilt_cookie *cookie, short *zoom);



RTV_DECL int pantilt_inq_power(pantilt_cookie *cookie, unsigned char *response);
/*response=0;on
  response=1;off*/


RTV_DECL int pantilt_inq_focusmode(pantilt_cookie *cookie,unsigned char *response);
/*response=0;auto
  response=1;manual*/


RTV_DECL int pantilt_inq_wbmode(pantilt_cookie *cookie,unsigned char *response);
/*response=0;auto
  response=1;indoor mode
  response=2;outdoor mode
  response=3;one push mode*/


RTV_DECL int pantilt_inq_aemode(pantilt_cookie *cookie,unsigned char *response);
/*	response=0;auto
	response=1;manual
	response=2;shutter priority
	response=3;Iris priority
	response=4;Bright Control*/


RTV_DECL int pantilt_inq_shutter(pantilt_cookie *cookie, unsigned short *shutter);// shutter:position, ZZZZ


RTV_DECL int pantilt_inq_iris(pantilt_cookie *cookie, unsigned short *iris);// shutter:position, ZZZZ
RTV_DECL int pantilt_inq_gain(pantilt_cookie *cookie, unsigned short *iris);// shutter:position, ZZZZ
RTV_DECL int pantilt_inq_backlight(pantilt_cookie *cookie,unsigned char *response);
/*response=0;on
  response=1;off*/


RTV_DECL int pantilt_inq_memory(pantilt_cookie *cookie,unsigned char *memory);//memory= 0 to 5


RTV_DECL int pantilt_inq_keylock(pantilt_cookie *cookie,unsigned char *response);
/*response=0;on
  response=1;off*/



RTV_DECL int pantilt_inq_id(pantilt_cookie *cookie, unsigned char *id);//ID= ZZ


RTV_DECL int pantilt_inq_receive(pantilt_cookie *cookie,unsigned char *response); 
/*response=0;on
  response=1;off*/



RTV_DECL int pantilt_inq_video(pantilt_cookie *cookie,unsigned char *response);
/*response=0;NTSC
  response=1;PAL*/



RTV_DECL int pantilt_inq_widelens(pantilt_cookie *cookie, unsigned char *lens); //lens= lens No Z


RTV_DECL int pantilt_inq_pantilt_mode(pantilt_cookie *cookie, unsigned char *status);//status=ZZZZ


RTV_DECL int pantilt_inq_pantilt_maxspeed(pantilt_cookie *cookie, unsigned char *pan_speed, unsigned char *tilt_speed);



RTV_DECL int pantilt_inq_datascreen(pantilt_cookie *cookie,unsigned char *response);
/*response=0;on
  response=1;off*/



RTV_DECL int pantilt_inq_atmd(pantilt_cookie *cookie,unsigned char *response);
/*response=0;normal mode
  response=1;AT mode
  response=2;MD mode*/
	


RTV_DECL int pantilt_inq_atmode(pantilt_cookie *cookie, unsigned char *status);//status=ZZZZ


RTV_DECL int pantilt_inq_atentry(pantilt_cookie *cookie,unsigned char *response) ;
/*response=0;entry mode 1 
  response=1;entry mode 2
  response=2;entry mode 3       
  response=3;entry mode 4*/


RTV_DECL int pantilt_inq_mdmode(pantilt_cookie *cookie, unsigned char *status);//status=ZZZZ


RTV_DECL int pantilt_inq_atposi(pantilt_cookie *cookie, unsigned char *xposition, unsigned char *yposition, unsigned char *status);
/* finds position of the object
   xposition=(04 to 2A)  yposition=(03 to 1B)
   status=ZZ  00(setting), 01(working), 10(lost a subject), 11(memorizing)*/

RTV_DECL int pantilt_inq_mdposi(pantilt_cookie *cookie, unsigned char *xposition, unsigned char *yposition, unsigned char *status);
/* motion detection object position inq
   xposition=(04 to 2A)  yposition=(03 to 1B)
   status=ZZ  00(setting), 01(undetected), 02(detecting), 3(memorizing)*/


RTV_DECL int pantilt_inq_mdylevel(pantilt_cookie *cookie, unsigned char *status); //status=ZZZZ


RTV_DECL int pantilt_inq_mdhuelevel(pantilt_cookie *cookie, unsigned char *status);//status=ZZZZ


RTV_DECL int pantilt_inq_mdsize(pantilt_cookie *cookie, unsigned char *status);//status=ZZZZ


RTV_DECL int pantilt_inq_md_disptime(pantilt_cookie *cookie, unsigned char *status);//status=ZZZZ


RTV_DECL int pantilt_inq_md_reftime(pantilt_cookie *cookie, unsigned char *status); 
/*response=0;refresh mode 1
  response=1;refresh mode 2
  response=2;refresh mode 3*/


/*MD_Ref.TimeInq*/
RTV_DECL int pantilt_inq_mdref(pantilt_cookie *cookie,unsigned char *response); //response=ZZZZ




/*test for the preence of the canera*/
RTV_DECL int presentcamera(pantilt_cookie *cookie);//returns 0 if camera is not connecte or OFF, or 1 if it can connect with the camera


RTV_DECL int pantilt_testcomm(pantilt_cookie *cookie);
/*call all the inquiries and makes a mess*/
RTV_DECL int monitor(pantilt_cookie *cookie);



/*Functions to be used only in case of emergency and total despair*/
/*do not use because it messes up the buffer since this command are not localized for only one camera*/ 
RTV_DECL int ifclear_broadcast(pantilt_cookie *cookie);
 /*this should work out ok, address= 1 to 7, call clearbuffer if you want to make inquiries after this call */
RTV_DECL int ifclear(pantilt_cookie *cookie, unsigned char address);
/*read everything that there is in the buffer, but it is a slow command since it waits a while after it stops recieving data*/
RTV_DECL int clearbuffer(pantilt_cookie *cookie);

#endif


/*made by Alexandre Saverin as158@cornell.edu
		  Pedro Felzenszwalb

  Vision Lab
  Cornell University
  Spring 98 
  Version 1.0*/
